""" Model the connection between a pipeline and a project or executor. """

import os
from collections.abc import Mapping
from logging import getLogger
from warnings import warn

import jsonschema
import pandas as pd
from eido import read_schema
from peppy import utils as peputil
from ubiquerg import expandpath, is_url
from yacman import load_yaml, YAMLConfigManager

from .const import *
from .exceptions import (
    InvalidResourceSpecificationException,
    PipelineInterfaceConfigError,
)
from .utils import jinja_render_template_strictly

__author__ = "Michal Stolarczyk"
__email__ = "michal@virginia.edu"

_LOGGER = getLogger(__name__)


@peputil.copy
class PipelineInterface(YAMLConfigManager):
    """
    This class parses, holds, and returns information for a yaml file that
    specifies how to interact with each individual pipeline. This
    includes both resources to request for cluster job submission, as well as
    arguments to be passed from the sample annotation metadata to the pipeline

    :param str | Mapping config: path to file from which to parse
        configuration data, or pre-parsed configuration data.
    :param str pipeline_type: type of the pipeline,
        must be either 'sample' or 'project'.
    """

    def __init__(self, config, pipeline_type=None):
        super(PipelineInterface, self).__init__()

        if isinstance(config, Mapping):
            self.pipe_iface_file = None
            self.source = None
        else:
            _LOGGER.debug("Reading {} from: {}".format(self.__class__.__name__, config))
            self.pipe_iface_file = config
            self.source = config
            config = load_yaml(config)
        if PIPELINE_INTERFACE_PIPELINE_NAME_KEY not in config:
            raise PipelineInterfaceConfigError(
                f"'{PIPELINE_INTERFACE_PIPELINE_NAME_KEY}' is required in pipeline interface config data."
            )
        self.update(config)
        self._validate(schema_src=PIFACE_SCHEMA_SRC)
        if "path" in self:
            warn(
                message="'path' specification as a top-level pipeline "
                "interface key is deprecated and will be removed with "
                "the next release. Please use 'paths' section "
                "from now on.",
                category=DeprecationWarning,
            )
            self._expand_paths(["path"])
        self._expand_paths(["compute", "dynamic_variables_script_path"])

    @property
    def pipeline_name(self):
        return self[PIPELINE_INTERFACE_PIPELINE_NAME_KEY]

    def render_var_templates(self, namespaces):
        """
        Render path templates under 'var_templates' in this pipeline interface.

        :param dict namespaces: namespaces to use for rendering
        """
        try:
            curr_data = self[VAR_TEMPL_KEY]
        except KeyError:
            _LOGGER.debug(
                f"'{VAR_TEMPL_KEY}' section not found in the "
                f"{self.__class__.__name__} object."
            )
            return {}
        else:
            var_templates = {}
            if curr_data:
                var_templates.update(curr_data)
                for k, v in var_templates.items():
                    var_templates[k] = jinja_render_template_strictly(v, namespaces)
            return var_templates

    def get_pipeline_schemas(self, schema_key=INPUT_SCHEMA_KEY):
        """
        Get path to the pipeline schema.

        :param str schema_key: where to look for schemas in the pipeline iface
        :return str: absolute path to the pipeline schema file
        """
        schema_source = None
        if schema_key in self:
            schema_source = self[schema_key]
        if schema_source:
            _LOGGER.debug("Got schema source: {}".format(schema_source))
            if is_url(schema_source):
                return schema_source
            elif not os.path.isabs(schema_source):
                schema_source = os.path.join(
                    os.path.dirname(self.pipe_iface_file), schema_source
                )
        return schema_source

    def choose_resource_package(self, namespaces, file_size):
        """
        Select resource bundle for given input file size to given pipeline.

        :param float file_size: Size of input data (in gigabytes).
        :param Mapping[Mapping[str]] namespaces: namespaced variables to pass
            as a context for fluid attributes command rendering
        :return MutableMapping: resource bundle appropriate for given pipeline,
            for given input file size
        :raises ValueError: if indicated file size is negative, or if the
            file size value specified for any resource package is negative
        :raises InvalidResourceSpecificationException: if no default
            resource package specification is provided
        """

        def _file_size_ante(name, data):
            # Retrieve this package's minimum file size.
            # Retain backwards compatibility while enforcing key presence.
            try:
                fsize = float(data[FILE_SIZE_COLNAME])
            except KeyError:
                raise InvalidResourceSpecificationException(
                    "Required column '{}' does not exist in resource "
                    "specification TSV.".format(FILE_SIZE_COLNAME)
                )
            # Negative file size is illogical and problematic for comparison.
            if fsize < 0:
                raise InvalidResourceSpecificationException(
                    "Found negative value () in '{}' column; package '{}'".format(
                        fsize, FILE_SIZE_COLNAME, name
                    )
                )
            return fsize

        def _notify(msg):
            msg += " for pipeline"
            if self.pipe_iface_file is not None:
                msg += " in interface {}".format(self.pipe_iface_file)
            _LOGGER.debug(msg)

        def _load_dynamic_vars(pipeline):
            """
            Render command string (jinja2 template), execute it in a subprocess
            and return its result (JSON object) as a dict

            :param Mapping pipeline: pipeline dict
            :return Mapping: a dict with attributes returned in the JSON
                by called command
            """

            def _log_raise_latest():
                """Log error info and raise latest handled exception"""
                _LOGGER.error(
                    "Could not retrieve JSON via command: '{}'".format(
                        pipeline[COMPUTE_KEY][DYN_VARS_KEY]
                    )
                )
                raise

            json = None
            if COMPUTE_KEY in pipeline and DYN_VARS_KEY in pipeline[COMPUTE_KEY]:
                from json import loads
                from subprocess import CalledProcessError, check_output
                from warnings import warn

                from .utils import jinja_render_template_strictly

                warn(
                    message="'dynamic_variables_command_template' feature is "
                    "deprecated and will be removed with the next "
                    "release. Please use 'pre_submit' feature from "
                    "now on.",
                    category=DeprecationWarning,
                )
                try:
                    cmd = jinja_render_template_strictly(
                        template=pipeline[COMPUTE_KEY][DYN_VARS_KEY],
                        namespaces=namespaces,
                    )
                    json = loads(check_output(cmd, shell=True))
                except CalledProcessError as e:
                    print(e.output)
                    _log_raise_latest()
                except Exception:
                    _log_raise_latest()
                else:
                    _LOGGER.debug(
                        "Loaded resources from JSON returned by a command for"
                        " pipeline '{}':\n{}".format(self.pipeline_name, json)
                    )
            return json

        def _load_size_dep_vars(piface):
            """
            Read the resources from a TSV provided in the pipeline interface

            :param looper.PipelineInterface piface: currently processed piface
            :param str section: section of pipeline interface to process
            :return pandas.DataFrame: resources
            """
            df = None
            if COMPUTE_KEY in piface and SIZE_DEP_VARS_KEY in piface[COMPUTE_KEY]:
                resources_tsv_path = piface[COMPUTE_KEY][SIZE_DEP_VARS_KEY]
                if not os.path.isabs(resources_tsv_path):
                    resources_tsv_path = os.path.join(
                        os.path.dirname(piface.pipe_iface_file), resources_tsv_path
                    )
                df = pd.read_csv(resources_tsv_path, sep="\t", header=0).fillna(
                    float("inf")
                )
                df[ID_COLNAME] = df.index
                df.set_index(ID_COLNAME)
                _LOGGER.debug(
                    "Loaded resources ({}) for pipeline '{}':\n{}".format(
                        resources_tsv_path, piface.pipeline_name, df
                    )
                )
            else:
                _notify("No '{}' defined".format(SIZE_DEP_VARS_KEY))
            return df

        # Ensure that we have a numeric value before attempting comparison.
        file_size = float(file_size)
        assert file_size >= 0, ValueError(
            "Attempted selection of resource "
            "package for negative file size: {}".format(file_size)
        )

        fluid_resources = _load_dynamic_vars(self)
        if fluid_resources is not None:
            return fluid_resources
        resources_df = _load_size_dep_vars(self)
        resources_data = {}
        if resources_df is not None:
            resources = resources_df.to_dict("index")
            try:
                # Sort packages by descending file size minimum to return first
                # package for which given file size satisfies the minimum.
                resource_packages = sorted(
                    resources.items(),
                    key=lambda name_and_data: _file_size_ante(*name_and_data),
                    reverse=False,
                )
            except ValueError:
                _LOGGER.error(
                    "Unable to use file size to prioritize "
                    "resource packages: {}".format(resources)
                )
                raise

            # choose minimally-sufficient package
            for rp_name, rp_data in resource_packages:
                size_ante = _file_size_ante(rp_name, rp_data)
                if file_size <= size_ante:
                    _LOGGER.debug(
                        "Selected '{}' package with file size {}Gb for file "
                        "of size {}Gb.".format(rp_name, size_ante, file_size)
                    )
                    _LOGGER.debug("Selected resource package data:\n{}".format(rp_data))
                    resources_data = rp_data
                    break

        if COMPUTE_KEY in self:
            resources_data.update(self[COMPUTE_KEY])
        project = namespaces["project"]
        if (
            LOOPER_KEY in project
            and COMPUTE_KEY in project[LOOPER_KEY]
            and RESOURCES_KEY in project[LOOPER_KEY][COMPUTE_KEY]
        ):
            # overwrite with values from project.looper.compute.resources
            resources_data.update(project[LOOPER_KEY][COMPUTE_KEY][RESOURCES_KEY])
        return resources_data

    def _expand_paths(self, keys):
        """
        Expand paths defined in the pipeline interface file

        :param list keys: list of keys resembling the nested structure to get
            to the pipeline interface attributre to expand
        """

        def _get_from_dict(map, attrs):
            """
            Get value from a possibly nested mapping using a list of its attributes

            :param collections.Mapping map: mapping to retrieve values from
            :param Iterable[str] attrs: a list of attributes
            :return: value found in the the requested attribute or
                None if one of the keys does not exist
            """
            for a in attrs:
                try:
                    map = map[a]
                except KeyError:
                    return
            return map

        def _set_in_dict(map, attrs, val):
            """
            Set value in a mapping, creating a possibly nested structure

            :param collections.Mapping map: mapping to retrieve values from
            :param Iterable[str] attrs: a list of attributes
            :param val: value to set
            :return: value found in the the requested attribute or
                None if one of the keys does not exist
            """
            for a in attrs:
                if a == attrs[-1]:
                    map[a] = val
                    break
                map.setdefault(a, PXAM())
                map = map[a]

        raw_path = _get_from_dict(self, keys)
        if not raw_path:
            return
        split_path = raw_path.split(" ")
        if len(split_path) > 1:
            _LOGGER.warning(
                "Path ({}) contains spaces. Using the first part as path: {}".format(
                    raw_path, split_path[0]
                )
            )
        path = split_path[0]
        pipe_path = expandpath(path)
        if not os.path.isabs(pipe_path) and self.pipe_iface_file:
            abs = os.path.join(os.path.dirname(self.pipe_iface_file), pipe_path)
            if os.path.exists(abs):
                _LOGGER.debug(
                    "Path relative to pipeline interface made absolute: {}".format(abs)
                )
                _set_in_dict(self, keys, abs)
                return
            _LOGGER.debug("Expanded path: {}".format(pipe_path))
            _set_in_dict(self, keys, pipe_path)

    def _validate(self, schema_src, exclude_case=False, flavor="generic"):
        """
        Generic function to validate the object against a schema

        :param str schema_src: schema source to validate against, URL or path
        :param bool exclude_case: whether to exclude validated objects
            from the error. Useful when used ith large projects
        :param str flavor: type of the pipeline schema to use
        """
        schema_source = schema_src.format(flavor)
        for schema in read_schema(schema_source):
            try:
                jsonschema.validate(self, schema)
                _LOGGER.debug(
                    f"Successfully validated {self.__class__.__name__} "
                    f"against schema: {schema_source}"
                )
            except jsonschema.exceptions.ValidationError as e:
                if not exclude_case:
                    raise e
                raise jsonschema.exceptions.ValidationError(e.message)



""" Looper version of NGS project model. """

import itertools
import os

try:
    from functools import cached_property
except ImportError:
    # cached_property was introduced in python 3.8
    cached_property = property
from logging import getLogger

from .divvy import ComputingConfiguration
from eido import PathAttrNotFoundError, read_schema
from jsonschema import ValidationError
from pandas.core.common import flatten
from peppy import CONFIG_KEY, OUTDIR_KEY
from peppy import Project as peppyProject
from peppy.utils import make_abs_via_cfg
from pipestat import PipestatError, PipestatManager
from ubiquerg import expandpath, is_command_callable
from yacman import YAMLConfigManager
from .conductor import write_pipestat_config

from .exceptions import *
from .pipeline_interface import PipelineInterface
from .processed_project import populate_project_paths, populate_sample_paths
from .utils import *

__all__ = ["Project"]

_LOGGER = getLogger(__name__)


class ProjectContext(object):
    """Wrap a Project to provide protocol-specific Sample selection."""

    def __init__(
        self,
        prj,
        selector_attribute=None,
        selector_include=None,
        selector_exclude=None,
        selector_flag=None,
        exclusion_flag=None,
    ):
        """Project and what to include/exclude defines the context."""
        if not isinstance(selector_attribute, str):
            raise TypeError(
                "Name of attribute for sample selection isn't a string: {} "
                "({})".format(selector_attribute, type(selector_attribute))
            )
        self.prj = prj
        self.include = selector_include
        self.exclude = selector_exclude
        self.attribute = selector_attribute
        self.selector_flag = selector_flag
        self.exclusion_flag = exclusion_flag

    def __getattr__(self, item):
        """Samples are context-specific; other requests are handled
        locally or dispatched to Project."""
        if item == "samples":
            return fetch_samples(
                prj=self.prj,
                selector_attribute=self.attribute,
                selector_include=self.include,
                selector_exclude=self.exclude,
                selector_flag=self.selector_flag,
                exclusion_flag=self.exclusion_flag,
            )
        if item in ["prj", "include", "exclude"]:
            # Attributes requests that this context/wrapper handles
            return self.__dict__[item]
        else:
            # Dispatch attribute request to Project.
            if hasattr(self.prj, item):
                return getattr(self.prj, item)
            else:
                return self.prj.get(item)

    def __getitem__(self, item):
        """Provide the Mapping-like item access to the instance's Project."""
        return self.prj[item]

    def __enter__(self):
        """References pass through this instance as needed, so the context
        provided is the instance itself."""
        return self

    def __repr__(self):
        return self.prj.__repr__()

    def __exit__(self, *args):
        """Context teardown."""
        pass


class Project(peppyProject):
    """
    Looper-specific Project.

    :param str cfg: path to configuration file with data from
        which Project is to be built
    :param Iterable[str] amendments: name indicating amendment to use, optional
    :param str divcfg_path: path to an environment configuration YAML file
        specifying compute settings.
    :param bool permissive: Whether a error should be thrown if
        a sample input file(s) do not exist or cannot be open.
    :param str compute_env_file: Environment configuration YAML file specifying
        compute settings.
    """

    def __init__(
        self, cfg=None, amendments=None, divcfg_path=None, runp=False, **kwargs
    ):
        super(Project, self).__init__(cfg=cfg, amendments=amendments)
        prj_dict = kwargs.get("project_dict")
        pep_config = kwargs.get("pep_config", None)
        if pep_config:
            self["pep_config"] = pep_config

        # init project from pephub pep_config:
        if prj_dict is not None and cfg is None:
            self.from_dict(prj_dict)
            self["_config_file"] = os.getcwd()  # for finding pipeline interface
            self["pep_config"] = pep_config

        self[EXTRA_KEY] = {}

        # add sample pipeline interface to the project
        if kwargs.get(SAMPLE_PL_ARG):
            self.set_sample_piface(kwargs.get(SAMPLE_PL_ARG))

        for attr_name in CLI_PROJ_ATTRS:
            if attr_name in kwargs:
                self[EXTRA_KEY][attr_name] = kwargs[attr_name]
                # setattr(self[EXTRA_KEY], attr_name, kwargs[attr_name])
        self._samples_by_interface = self._samples_by_piface(self.piface_key)
        self._interfaces_by_sample = self._piface_by_samples()
        self.linked_sample_interfaces = self._get_linked_pifaces()
        if FILE_CHECKS_KEY in self[EXTRA_KEY]:
            setattr(self, "file_checks", not self[EXTRA_KEY][FILE_CHECKS_KEY])
        if DRY_RUN_KEY in self[EXTRA_KEY]:
            setattr(self, DRY_RUN_KEY, self[EXTRA_KEY][DRY_RUN_KEY])
        self.dcc = (
            None
            if divcfg_path is None
            else ComputingConfiguration(filepath=divcfg_path)
        )
        if DRY_RUN_KEY in self and not self[DRY_RUN_KEY]:
            _LOGGER.debug("Ensuring project directories exist")
            self.make_project_dirs()

    @property
    def piface_key(self):
        """
        Name of the pipeline interface attribute for this project

        :return str: name of the pipeline interface attribute
        """
        return self._extra_cli_or_cfg(PIFACE_KEY_SELECTOR) or PIPELINE_INTERFACES_KEY

    @property
    def selected_compute_package(self):
        """
        Compute package name specified in object constructor

        :return str: compute package name
        """
        return self._extra_cli_or_cfg(COMPUTE_PACKAGE_KEY)

    @property
    def cli_pifaces(self):
        """
        Collection of pipeline interface sources specified in object constructor

        :return list[str]: collection of pipeline interface sources
        """
        x = self._extra_cli_or_cfg(self.piface_key)
        return (
            list(flatten([x] if not isinstance(x, list) else x))
            if x is not None
            else None
        )

    @property
    def output_dir(self):
        """
        Output directory for the project, specified in object constructor

        :return str: path to the output directory
        """
        return self._extra_cli_or_cfg(OUTDIR_KEY, strict=True)

    def _extra_cli_or_cfg(self, attr_name, strict=False):
        """
        Get attribute value provided in kwargs in object constructor of from
        looper section in the configuration file

        :param str attr_name: name of the attribute to get value for
        :param bool strict: whether a non-existent attribute is exceptional
        :raise MisconfigurationException: in strict mode, when no attribute
         found
        """
        try:
            result = self[EXTRA_KEY][attr_name]
            # getattr(self[EXTRA_KEY], attr_name))
        except (AttributeError, KeyError):
            pass
        else:
            if result is not None:
                return result
        if (
            CONFIG_KEY in self
            and LOOPER_KEY in self[CONFIG_KEY]
            and attr_name in self[CONFIG_KEY][LOOPER_KEY]
        ):
            return self[CONFIG_KEY][LOOPER_KEY][attr_name]
        else:
            if strict:
                raise MisconfigurationException(
                    "'{}' is missing. Provide it in the '{}' section of the "
                    "project configuration file".format(attr_name, LOOPER_KEY)
                )
            return

    @property
    def results_folder(self):
        """
        Path to the results folder for the project

        :return str: path to the results folder in the output folder
        """
        return self._out_subdir_path(RESULTS_SUBDIR_KEY, default="results_pipeline")

    @property
    def submission_folder(self):
        """
        Path to the submission folder for the project

        :return str: path to the submission in the output folder
        """
        return self._out_subdir_path(SUBMISSION_SUBDIR_KEY, default="submission")

    def _out_subdir_path(self, key: str, default: str) -> str:
        """
        Create a system path relative to the project output directory.
        The values for the names of the subdirectories are sourced from
        kwargs passed to the object constructor.

        :param str key: name of the attribute mapped to the value of interest
        :param str default: if key not specified, a default to use
        :return str: path to the folder
        """
        parent = getattr(self, OUTDIR_KEY)
        child = getattr(self[EXTRA_KEY], key, default) or default
        return os.path.join(parent, child)

    def make_project_dirs(self):
        """
        Create project directory structure if it doesn't exist.
        """
        for folder_key in ["results_folder", "submission_folder"]:
            folder_path = getattr(self, folder_key)
            _LOGGER.debug("Ensuring project dir exists: '{}'".format(folder_path))
            if not os.path.exists(folder_path):
                _LOGGER.debug(
                    "Attempting to create project folder: '{}'".format(folder_path)
                )
                try:
                    os.makedirs(folder_path)
                except OSError as e:
                    _LOGGER.warning(
                        "Could not create project folder: '{}'".format(str(e))
                    )

    @cached_property
    def project_pipeline_interface_sources(self):
        """
        Get a list of all valid project-level pipeline interface sources
        associated with this project. Sources that are file paths are expanded

        :return list[str]: collection of valid pipeline interface sources:
        """
        return (
            [self._resolve_path_with_cfg(src) for src in self.cli_pifaces]
            if self.cli_pifaces is not None
            else []
        )

    @cached_property
    def project_pipeline_interfaces(self):
        """
        Flat list of all valid project-level interface objects associated
        with this Project

        Note that only valid pipeline interfaces will show up in the
        result (ones that exist on disk/remotely and validate successfully
        against the schema)

        :return list[looper.PipelineInterface]: list of pipeline interfaces
        """
        return [
            PipelineInterface(pi, pipeline_type="project")
            for pi in self.project_pipeline_interface_sources
        ]

    @cached_property
    def pipeline_interfaces(self):
        """
        Flat list of all valid interface objects associated with this Project

        Note that only valid pipeline interfaces will show up in the
        result (ones that exist on disk/remotely and validate successfully
        against the schema)

        :return list[looper.PipelineInterface]: list of pipeline interfaces
        """
        return [pi for ifaces in self._interfaces_by_sample.values() for pi in ifaces]

    @cached_property
    def pipeline_interface_sources(self):
        """
        Get a list of all valid pipeline interface sources associated
        with this project. Sources that are file paths are expanded

        :return list[str]: collection of valid pipeline interface sources
        """
        return self._samples_by_interface.keys()

    @cached_property
    def pipestat_configured(self):
        """
        Whether pipestat configuration is complete for all sample pipelines

        :return bool: whether pipestat configuration is complete
        """
        return self._check_if_pipestat_configured()

    @cached_property
    def pipestat_configured_project(self):
        """
        Whether pipestat configuration is complete for all project pipelines

        :return bool: whether pipestat configuration is complete
        """
        return self._check_if_pipestat_configured(project_level=True)

    def get_sample_piface(self, sample_name):
        """
        Get a list of pipeline interfaces associated with the specified sample.

        Note that only valid pipeline interfaces will show up in the
        result (ones that exist on disk/remotely and validate successfully
        against the schema)

        :param str sample_name: name of the sample to retrieve list of
            pipeline interfaces for
        :return list[looper.PipelineInterface]: collection of valid
            pipeline interfaces associated with selected sample
        """
        try:
            return self._interfaces_by_sample[sample_name]
        except KeyError:
            return None

    def build_submission_bundles(self, protocol, priority=True):
        """
        Create pipelines to submit for each sample of a particular protocol.

        With the argument (flag) to the priority parameter, there's control
        over whether to submit pipeline(s) from only one of the project's
        known pipeline locations with a match for the protocol, or whether to
        submit pipelines created from all locations with a match for the
        protocol.

        :param str protocol: name of the protocol/library for which to
            create pipeline(s)
        :param bool priority: to only submit pipeline(s) from the first of the
            pipelines location(s) (indicated in the project config file) that
            has a match for the given protocol; optional, default True
        :return Iterable[(PipelineInterface, type, str, str)]:
        :raises AssertionError: if there's a failure in the attempt to
            partition an interface's pipeline scripts into disjoint subsets of
            those already mapped and those not yet mapped
        """

        if not priority:
            raise NotImplementedError(
                "Currently, only prioritized protocol mapping is supported "
                "(i.e., pipeline interfaces collection is a prioritized list, "
                "so only the first interface with a protocol match is used.)"
            )

        # Pull out the collection of interfaces (potentially one from each of
        # the locations indicated in the project configuration file) as a
        # sort of pool of information about possible ways in which to submit
        # pipeline(s) for sample(s) of the indicated protocol.
        pifaces = self.interfaces.get_pipeline_interface(protocol)
        if not pifaces:
            raise PipelineInterfaceConfigError(
                "No interfaces for protocol: {}".format(protocol)
            )

        # coonvert to a list, in the future we might allow to match multiple
        pifaces = pifaces if isinstance(pifaces, str) else [pifaces]

        job_submission_bundles = []
        new_jobs = []

        _LOGGER.debug("Building pipelines matched by protocol: {}".format(protocol))

        for pipe_iface in pifaces:
            # Determine how to reference the pipeline and where it is.
            path = pipe_iface["path"]
            if not (os.path.exists(path) or is_command_callable(path)):
                _LOGGER.warning("Missing pipeline script: {}".format(path))
                continue

            # Add this bundle to the collection of ones relevant for the
            # current PipelineInterface.
            new_jobs.append(pipe_iface)
            job_submission_bundles.append(new_jobs)
        return list(itertools.chain(*job_submission_bundles))

    @staticmethod
    def get_schemas(pifaces, schema_key=INPUT_SCHEMA_KEY):
        """
        Get the list of unique schema paths for a list of pipeline interfaces

        :param str | Iterable[str] pifaces: pipeline interfaces to search
            schemas for
        :param str schema_key: where to look for schemas in the piface
        :return Iterable[str]: unique list of schema file paths
        """
        if isinstance(pifaces, str):
            pifaces = [pifaces]
        schema_set = set()
        for piface in pifaces:
            schema_file = piface.get_pipeline_schemas(schema_key)
            if schema_file:
                schema_set.update([schema_file])
        return list(schema_set)

    def get_pipestat_managers(self, sample_name=None, project_level=False):
        """
        Get a collection of pipestat managers for the selected sample or project.

        The number of pipestat managers corresponds to the number of unique
        output schemas in the pipeline interfaces specified by the sample or project.

        :param str sample_name: sample name to get pipestat managers for
        :param bool project_level: whether the project PipestatManagers
            should be returned
        :return dict[str, pipestat.PipestatManager]: a mapping of pipestat
            managers by pipeline interface name
        """
        pipestat_configs = self._get_pipestat_configuration(
            sample_name=sample_name, project_level=project_level
        )
        return {
            pipeline_name: PipestatManager(**pipestat_vars)
            for pipeline_name, pipestat_vars in pipestat_configs.items()
        }

    def _check_if_pipestat_configured(self, project_level=False):
        """
        A helper method determining whether pipestat configuration is complete

        :param bool project_level: whether the project pipestat config should be checked
        :return bool: whether pipestat configuration is complete
        """
        try:
            if project_level:
                pipestat_configured = self._get_pipestat_configuration(
                    sample_name=None, project_level=project_level
                )
            else:
                for s in self.samples:
                    pipestat_configured = self._get_pipestat_configuration(
                        sample_name=s.sample_name
                    )
        except Exception as e:
            context = (
                f"Project '{self.name}'"
                if project_level
                else f"Sample '{s.sample_name}'"
            )
            _LOGGER.debug(
                f"Pipestat configuration incomplete for {context}; "
                f"caught exception: {getattr(e, 'message', repr(e))}"
            )
            return False
        else:
            if pipestat_configured is not None and pipestat_configured != {}:
                return True
            else:
                return False

    def _get_pipestat_configuration(self, sample_name=None, project_level=False):
        """
        Get all required pipestat configuration variables from looper_config file
        """

        ret = {}
        if not project_level and sample_name is None:
            raise ValueError(
                "Must provide the sample_name to determine the "
                "sample to get the PipestatManagers for"
            )

        if PIPESTAT_KEY in self[EXTRA_KEY]:
            pipestat_config_dict = self[EXTRA_KEY][PIPESTAT_KEY]
        else:
            _LOGGER.debug(
                f"'{PIPESTAT_KEY}' not found in '{LOOPER_KEY}' section of the "
                f"project configuration file."
            )
            # We cannot use pipestat without it being defined in the looper config file.
            raise ValueError

        # Expand paths in the event ENV variables were used in config files
        output_dir = expandpath(self.output_dir)

        # Get looper user configured items first and update the pipestat_config_dict
        try:
            results_file_path = expandpath(pipestat_config_dict["results_file_path"])
            if not os.path.exists(os.path.dirname(results_file_path)):
                results_file_path = os.path.join(
                    os.path.dirname(output_dir), results_file_path
                )
            pipestat_config_dict.update({"results_file_path": results_file_path})
        except KeyError:
            results_file_path = None

        try:
            flag_file_dir = expandpath(pipestat_config_dict["flag_file_dir"])
            if not os.path.isabs(flag_file_dir):
                flag_file_dir = os.path.join(os.path.dirname(output_dir), flag_file_dir)
            pipestat_config_dict.update({"flag_file_dir": flag_file_dir})
        except KeyError:
            flag_file_dir = None

        if sample_name:
            pipestat_config_dict.update({"record_identifier": sample_name})

        if project_level and "project_name" in pipestat_config_dict:
            pipestat_config_dict.update(
                {"project_name": pipestat_config_dict["project_name"]}
            )

        if project_level and "{record_identifier}" in results_file_path:
            # if project level and using {record_identifier}, pipestat needs some sort of record_identifier during creation
            pipestat_config_dict.update(
                {"record_identifier": "default_project_record_identifier"}
            )

        pipestat_config_dict.update({"output_dir": output_dir})

        pifaces = (
            self.project_pipeline_interfaces
            if project_level
            else self._interfaces_by_sample[sample_name]
        )

        for piface in pifaces:
            # We must also obtain additional pipestat items from the pipeline author's piface
            if "output_schema" in piface.data:
                schema_path = expandpath(piface.data["output_schema"])
                if not os.path.isabs(schema_path):
                    # Get path relative to the pipeline_interface
                    schema_path = os.path.join(
                        os.path.dirname(piface.pipe_iface_file), schema_path
                    )
                pipestat_config_dict.update({"schema_path": schema_path})
            if "pipeline_name" in piface.data:
                pipestat_config_dict.update(
                    {"pipeline_name": piface.data["pipeline_name"]}
                )
            if "pipeline_type" in piface.data:
                pipestat_config_dict.update(
                    {"pipeline_type": piface.data["pipeline_type"]}
                )

            # Pipestat_dict_ is now updated from all sources and can be written to a yaml.
            looper_pipestat_config_path = os.path.join(
                os.path.dirname(output_dir), "looper_pipestat_config.yaml"
            )
            write_pipestat_config(looper_pipestat_config_path, pipestat_config_dict)

            ret[piface.pipeline_name] = {
                "config_file": looper_pipestat_config_path,
            }
        return ret

    def populate_pipeline_outputs(self):
        """
        Populate project and sample output attributes based on output schemas
        that pipeline interfaces point to.
        """
        # eido.read_schema always returns a list of schemas since it supports
        # imports in schemas. The output schemas can't have the import section,
        # hence it's safe to select the fist element after read_schema() call.
        for sample in self.samples:
            sample_piface = self.get_sample_piface(sample[self.sample_table_index])
            if sample_piface:
                paths = self.get_schemas(sample_piface, OUTPUT_SCHEMA_KEY)
                for path in paths:
                    populate_sample_paths(sample, read_schema(path)[0])
        schemas = self.get_schemas(self.project_pipeline_interfaces, OUTPUT_SCHEMA_KEY)
        for schema in schemas:
            populate_project_paths(self, read_schema(schema)[0])

    def _get_linked_pifaces(self):
        """
        Get linked sample pipeline interfaces by project pipeline interface.

        These are indicated in project pipeline interface by
        'linked_pipeline_interfaces' key. If a project pipeline interface
         does not have such key defined, an empty list is returned for that
         pipeline interface.

        :return dict[list[str]]: mapping of sample pipeline interfaces
            by project pipeline interfaces
        """

        def _process_linked_piface(p, piface, prj_piface):
            piface = make_abs_via_cfg(piface, prj_piface)
            if piface not in p.pipeline_interface_sources:
                raise PipelineInterfaceConfigError(
                    "Linked sample pipeline interface was not assigned "
                    f"to any sample in this project: {piface}"
                )
            return piface

        linked_pifaces = {}
        for prj_piface in self.project_pipeline_interfaces:
            pifaces = (
                prj_piface.linked_pipeline_interfaces
                if hasattr(prj_piface, "linked_pipeline_interfaces")
                else []
            )
            linked_pifaces[prj_piface.source] = list(
                {
                    _process_linked_piface(self, piface, prj_piface.source)
                    for piface in pifaces
                }
            )
        return linked_pifaces

    def _piface_by_samples(self):
        """
        Create a mapping of all defined interfaces in this Project by samples.

        :return dict[str, list[PipelineInterface]]: a collection of pipeline
            interfaces keyed by sample name
        """
        pifaces_by_sample = {}
        for source, sample_names in self._samples_by_interface.items():
            try:
                pi = PipelineInterface(source, pipeline_type="sample")
            except PipelineInterfaceConfigError as e:
                _LOGGER.debug(f"Skipping pipeline interface creation: {e}")
            else:
                for sample_name in sample_names:
                    pifaces_by_sample.setdefault(sample_name, []).append(pi)
        return pifaces_by_sample

    def _omit_from_repr(self, k, cls):
        """
        Exclude the interfaces from representation.

        :param str k: key of item to consider for omission
        :param type cls: placeholder to comply with superclass signature
        """
        return super(Project, self)._omit_from_repr(k, cls) or k == "interfaces"

    def _resolve_path_with_cfg(self, pth):
        """
        Expand provided path and make it absolute using project config path

        :param str pth: path, possibly including env vars and/or relative
        :return str: absolute path
        """
        if pth is None:
            return
        pth = expandpath(pth)
        if not os.path.isabs(pth):
            pth = os.path.realpath(os.path.join(os.path.dirname(self.config_file), pth))
            _LOGGER.debug("Relative path made absolute: {}".format(pth))
        return pth

    def _samples_by_piface(self, piface_key):
        """
        Create a collection of all samples with valid pipeline interfaces

        :param str piface_key: name of the attribute that holds pipeline
         interfaces
        :return list[str]: a collection of samples keyed by pipeline interface
            source
        """
        samples_by_piface = {}
        msgs = set()
        for sample in self.samples:
            if piface_key in sample and sample[piface_key]:
                piface_srcs = sample[piface_key]
                if isinstance(piface_srcs, str):
                    piface_srcs = [piface_srcs]
                for source in piface_srcs:
                    source = self._resolve_path_with_cfg(source)
                    try:
                        PipelineInterface(source, pipeline_type="sample")
                    except (
                        ValidationError,
                        IOError,
                        PipelineInterfaceConfigError,
                    ) as e:
                        msg = (
                            "Ignoring invalid pipeline interface source: "
                            "{}. Caught exception: {}".format(
                                source, getattr(e, "message", repr(e))
                            )
                        )
                        msgs.add(msg)
                        continue
                    else:
                        samples_by_piface.setdefault(source, set())
                        samples_by_piface[source].add(sample[self.sample_table_index])
        for msg in msgs:
            _LOGGER.warning(msg)
        return samples_by_piface

    def set_sample_piface(self, sample_piface: Union[List[str], str]) -> NoReturn:
        """
        Add sample pipeline interfaces variable to object

        :param list | str sample_piface: sample pipeline interface
        """
        self.config.setdefault("sample_modifiers", {})
        self.config["sample_modifiers"].setdefault("append", {})
        self.config["sample_modifiers"]["append"]["pipeline_interfaces"] = sample_piface

        self.modify_samples()


def fetch_samples(
    prj,
    selector_attribute=None,
    selector_include=None,
    selector_exclude=None,
    selector_flag=None,
    exclusion_flag=None,
):
    """
    Collect samples of particular protocol(s).

    Protocols can't be both positively selected for and negatively
    selected against. That is, it makes no sense and is not allowed to
    specify both selector_include and selector_exclude protocols. On the
    other hand, if
    neither is provided, all of the Project's Samples are returned.
    If selector_include is specified, Samples without a protocol will be
    excluded,
    but if selector_exclude is specified, protocol-less Samples will be
    included.

    :param Project prj: the Project with Samples to fetch
    :param str selector_attribute: name of attribute on which to base the
    fetch
    :param Iterable[str] | str selector_include: protocol(s) of interest;
        if specified, a Sample must
    :param Iterable[str] | str selector_exclude: protocol(s) to include
    :param Iterable[str] | str selector_flag: flag to select on, e.g. FAILED, COMPLETED
    :param Iterable[str] | str exclusion_flag: flag to exclude on, e.g. FAILED, COMPLETED
    :return list[Sample]: Collection of this Project's samples with
        protocol that either matches one of those in selector_include,
        or either
        lacks a protocol or does not match one of those in selector_exclude
    :raise TypeError: if both selector_include and selector_exclude
    protocols are
        specified; TypeError since it's basically providing two arguments
        when only one is accepted, so remain consistent with vanilla
        Python2;
        also possible if name of attribute for selection isn't a string
    """

    kept_samples = prj.samples

    if not selector_include and not selector_exclude:
        # Default case where user does not use selector_include or selector exclude.
        # Assume that user wants to exclude samples if toggle = 0.
        # if any([hasattr(s, "toggle") for s in prj.samples]):
        # if any("toggle" in s for s in prj.samples):
        if "toggle" in prj.samples[0]:  # assume the samples have the same schema
            selector_exclude = [0]

            def keep(s):
                return (
                    not hasattr(s, selector_attribute)
                    or getattr(s, selector_attribute) not in selector_exclude
                )

            kept_samples = list(filter(keep, prj.samples))
        else:
            kept_samples = prj.samples

    # Intersection between selector_include and selector_exclude is
    # nonsense user error.
    if selector_include and selector_exclude:
        raise TypeError(
            "Specify only selector_include or selector_exclude parameter, " "not both."
        )

    if not isinstance(selector_attribute, str):
        raise TypeError(
            "Name for attribute on which to base selection isn't string: "
            "{} "
            "({})".format(selector_attribute, type(selector_attribute))
        )

    # At least one of the samples has to have the specified attribute
    if prj.samples and not any([hasattr(s, selector_attribute) for s in prj.samples]):
        if selector_attribute == "toggle":
            # this is the default, so silently pass.
            pass
        else:
            raise AttributeError(
                "The Project samples do not have the attribute '{attr}'".format(
                    attr=selector_attribute
                )
            )

    if prj.samples:
        # Use the attr check here rather than exception block in case the
        # hypothetical AttributeError would occur; we want such
        # an exception to arise, not to catch it as if the Sample lacks
        # "protocol"
        if not selector_include:
            # Loose; keep all samples not in the selector_exclude.
            def keep(s):
                return not hasattr(s, selector_attribute) or getattr(
                    s, selector_attribute
                ) not in make_set(selector_exclude)

        else:
            # Strict; keep only samples in the selector_include.
            def keep(s):
                return hasattr(s, selector_attribute) and getattr(
                    s, selector_attribute
                ) in make_set(selector_include)

        kept_samples = list(filter(keep, kept_samples))

        if selector_flag and exclusion_flag:
            raise TypeError("Specify only selector_flag or exclusion_flag not both.")

        flags = selector_flag or exclusion_flag or None
        if flags:
            # Collect uppercase flags or error if not str
            if not isinstance(flags, list):
                flags = [str(flags)]
            for flag in flags:
                if not isinstance(flag, str):
                    raise TypeError(
                        f"Supplied flags must be a string! Flag:{flag} {type(flag)}"
                    )
                flags.remove(flag)
                flags.insert(0, flag.upper())
            # Look for flags
            # Is pipestat configured? Then, the user may have set the flag folder
            if prj.pipestat_configured:
                try:
                    flag_dir = expandpath(prj[EXTRA_KEY][PIPESTAT_KEY]["flag_file_dir"])
                    if not os.path.isabs(flag_dir):
                        flag_dir = os.path.join(
                            os.path.dirname(prj.output_dir), flag_dir
                        )
                except KeyError:
                    _LOGGER.warning(
                        "Pipestat is configured but no flag_file_dir supplied, defaulting to output_dir"
                    )
                    flag_dir = prj.output_dir
            else:
                # if pipestat not configured, check the looper output dir
                flag_dir = prj.output_dir

            # Using flag_dir, search for flags:
            for sample in kept_samples:
                sample_pifaces = prj.get_sample_piface(sample[prj.sample_table_index])
                pl_name = sample_pifaces[0].pipeline_name
                flag_files = fetch_sample_flags(prj, sample, pl_name, flag_dir)
                status = get_sample_status(sample.sample_name, flag_files)
                sample.update({"status": status})

            if not selector_flag:
                # Loose; keep all samples not in the exclusion_flag.
                def keep(s):
                    return not hasattr(s, "status") or getattr(
                        s, "status"
                    ) not in make_set(flags)

            else:
                # Strict; keep only samples in the selector_flag
                def keep(s):
                    return hasattr(s, "status") and getattr(s, "status") in make_set(
                        flags
                    )

            kept_samples = list(filter(keep, kept_samples))

    return kept_samples


def make_set(items):
    try:
        # Check if user input single integer value for inclusion/exclusion criteria
        if len(items) == 1:
            items = list(map(int, items))  # list(int(items[0]))
    except:
        if isinstance(items, str):
            items = [items]
    return items



""" Computing configuration representation """

import logging
import logmuse
import os
import sys
import shutil
import yaml
from yaml import SafeLoader
from shutil import copytree

from ubiquerg import is_writable, VersionInHelpParser
import yacman

from .const import (
    COMPUTE_SETTINGS_VARNAME,
    DEFAULT_COMPUTE_RESOURCES_NAME,
    NEW_COMPUTE_KEY,
    DEFAULT_CONFIG_FILEPATH,
    DEFAULT_CONFIG_SCHEMA,
)
from .utils import write_submit_script

# from . import __version__

_LOGGER = logging.getLogger(__name__)

# This is the divvy.py submodule from divvy


class ComputingConfiguration(yacman.YAMLConfigManager):
    """
    Represents computing configuration objects.

    The ComputingConfiguration class provides a computing configuration object
    that is an *in memory* representation of a `divvy` computing configuration
    file. This object has various functions to allow a user to activate, modify,
    and retrieve computing configuration files, and use these values to populate
    job submission script templates.

    :param str | Iterable[(str, object)] | Mapping[str, object] entries: config
        Collection of key-value pairs.
    :param str filepath: YAML file specifying computing package data. (the
        `DIVCFG` file)
    """

    def __init__(self, entries=None, filepath=None):
        if not entries and not filepath:
            # Handle the case of an empty one, when we'll use the default
            filepath = select_divvy_config(None)

        super(ComputingConfiguration, self).__init__(
            entries=entries,
            filepath=filepath,
            schema_source=DEFAULT_CONFIG_SCHEMA,
            validate_on_write=True,
        )

        if not "compute_packages" in self:
            raise Exception(
                "Your divvy config file is not in divvy config format "
                "(it lacks a compute_packages section): '{}'".format(filepath)
            )
            # We require that compute_packages be present, even if empty
            self["compute_packages"] = {}

        # Initialize default compute settings.
        _LOGGER.debug("Establishing project compute settings")
        self.compute = None
        self.setdefault("adapters", None)
        self.activate_package(DEFAULT_COMPUTE_RESOURCES_NAME)
        self.config_file = self.filepath

    def write(self, filename=None):
        super(ComputingConfiguration, self).write(filepath=filename, exclude_case=True)
        filename = filename or getattr(self, yacman.FILEPATH_KEY)
        filedir = os.path.dirname(filename)
        # For this object, we *also* have to write the template files
        for pkg_name, pkg in self["compute_packages"].items():
            print(pkg)
            destfile = os.path.join(filedir, os.path.basename(pkg.submission_template))
            shutil.copyfile(pkg.submission_template, destfile)

    @property
    def compute_env_var(self):
        """
        Environment variable through which to access compute settings.

        :return list[str]: names of candidate environment variables, for which
            value may be path to compute settings file; first found is used.
        """
        return COMPUTE_SETTINGS_VARNAME

    @property
    def default_config_file(self):
        """
        Path to default compute environment settings file.

        :return str: Path to default compute settings file
        """
        return DEFAULT_CONFIG_FILEPATH

    # Warning: template cannot be a property, because otherwise
    # it will get treated as a PathExAttMap treats all properties, which
    # is that it will turn any double-slashes into single slashes.
    def template(self):
        """
        Get the currently active submission template.

        :return str: submission script content template for current state
        """
        with open(self.compute["submission_template"], "r") as f:
            return f.read()

    @property
    def templates_folder(self):
        """
        Path to folder with default submission templates.

        :return str: path to folder with default submission templates
        """
        return os.path.join(
            os.path.dirname(__file__), "default_config", "divvy_templates"
        )

    def activate_package(self, package_name):
        """
        Activates a compute package.

        This copies the computing attributes from the configuration file into
        the `compute` attribute, where the class stores current compute
        settings.

        :param str package_name: name for non-resource compute bundle,
            the name of a subsection in an environment configuration file
        :return bool: success flag for attempt to establish compute settings
        """

        # Hope that environment & environment compute are present.
        act_msg = "Activating compute package '{}'".format(package_name)
        if package_name == "default":
            _LOGGER.debug(act_msg)
        else:
            _LOGGER.info(act_msg)

        if (
            package_name
            and self["compute_packages"]
            and package_name in self["compute_packages"]
        ):
            # Augment compute, creating it if needed.
            if self.compute is None:
                _LOGGER.debug("Creating Project compute")
                self.compute = yacman.YAMLConfigManager()
                _LOGGER.debug(
                    "Adding entries for package_name '{}'".format(package_name)
                )

            self.compute.update(self["compute_packages"][package_name])

            # Ensure submission template is absolute. This *used to be* handled
            # at update (so the paths were stored as absolutes in the packages),
            # but now, it makes more sense to do it here so we can piggyback on
            # the default update() method and not even have to do that.
            if not os.path.isabs(self.compute["submission_template"]):
                try:
                    self.compute["submission_template"] = os.path.join(
                        os.path.dirname(self.filepath),
                        self.compute["submission_template"],
                    )
                except AttributeError as e:
                    # Environment and environment compute should at least have been
                    # set as null-valued attributes, so execution here is an error.
                    _LOGGER.error(str(e))

            _LOGGER.debug(
                "Submit template set to: {}".format(self.compute["submission_template"])
            )

            return True

        else:
            # Scenario in which environment and environment compute are
            # both present--but don't evaluate to True--is fairly harmless.
            _LOGGER.debug(
                "Can't activate package. compute_packages = {}".format(
                    self["compute_packages"]
                )
            )

        return False

    def clean_start(self, package_name):
        """
        Clear current active settings and then activate the given package.

        :param str package_name: name of the resource package to activate
        :return bool: success flag
        """
        self.reset_active_settings()
        return self.activate_package(package_name)

    def get_active_package(self):
        """
        Returns settings for the currently active compute package

        :return yacman.YacAttMap: data defining the active compute package
        """
        return self.compute

    def list_compute_packages(self):
        """
        Returns a list of available compute packages.

        :return set[str]: names of available compute packages
        """
        return set(self["compute_packages"].keys())

    def reset_active_settings(self):
        """
        Clear out current compute settings.

        :return bool: success flag
        """
        self.compute = yacman.YacAttMap()
        return True

    def update_packages(self, config_file):
        """
        Parse data from divvy configuration file.

        Given a divvy configuration file, this function will update (not
        overwrite) existing compute packages with existing values. It does not
        affect any currently active settings.

        :param str config_file: path to file with new divvy configuration data
        """
        entries = yacman.load_yaml(config_file)
        self.update(entries)
        return True

    def get_adapters(self):
        """
        Get current adapters, if defined.

        Adapters are sourced from the 'adapters' section in the root of the
        divvy configuration file and updated with an active compute
        package-specific set of adapters, if any defined in 'adapters' section
        under currently active compute package.

        :return yacman.YAMLConfigManager: current adapters mapping
        """
        adapters = yacman.YAMLConfigManager()
        if "adapters" in self and self["adapters"] is not None:
            adapters.update(self["adapters"])
        if "compute" in self and "adapters" in self.compute:
            adapters.update(self.compute["adapters"])
        if not adapters:
            _LOGGER.debug("No adapters determined in divvy configuration file.")
        return adapters

    def submit(self, output_path, extra_vars=None):
        if not output_path:
            import tempfile

            with tempfile.NamedTemporaryFile() as temp:
                _LOGGER.info(
                    "No file provided; using temp file: '{}'".format(temp.name)
                )
                self.submit(temp.name, extra_vars)
        else:
            script = self.write_script(output_path, extra_vars)
            submission_command = "{} {}".format(
                self.compute["submission_command"], script
            )
            _LOGGER.info(submission_command)
            os.system(submission_command)

    def write_script(self, output_path, extra_vars=None):
        """
        Given currently active settings, populate the active template to write a
         submission script. Additionally use the current adapters to adjust
         the select of the provided variables

        :param str output_path: Path to file to write as submission script
        :param Iterable[Mapping] extra_vars: A list of Dict objects with
            key-value pairs with which to populate template fields. These will
            override any values in the currently active compute package.
        :return str: Path to the submission script file
        """

        def _get_from_dict(map, attrs):
            """
            Get value from a possibly mapping using a list of its attributes

            :param collections.Mapping map: mapping to retrieve values from
            :param Iterable[str] attrs: a list of attributes
            :return: value found in the the requested attribute or
                None if one of the keys does not exist
            """
            for a in attrs:
                try:
                    map = map[a]
                except KeyError:
                    return None
            return map

        from copy import deepcopy

        variables = deepcopy(self.compute)
        _LOGGER.debug("Extra vars: {}".format(extra_vars))
        if extra_vars:
            if not isinstance(extra_vars, list):
                extra_vars = [extra_vars]
            adapters = self.get_adapters()
            exclude = set()
            if adapters:
                # apply adapted values first and keep track of
                # which of extra_vars were used
                for n, v in adapters.items():
                    split_v = v.split(".")
                    namespace = split_v[0]
                    for extra_var in reversed(extra_vars):
                        if (
                            len(extra_var) > 0
                            and namespace in list(extra_var.keys())[0]
                        ):
                            exclude.add(namespace)
                            var = _get_from_dict(extra_var, split_v)
                            if var is not None:
                                variables[n] = var
                                _LOGGER.debug(
                                    "adapted {}: ({}={})".format(
                                        n, ".".join(split_v), var
                                    )
                                )
            for extra_var in reversed(extra_vars):
                # then update variables with the rest of the extra_vars
                if len(extra_var) > 0 and list(extra_var.keys())[0] not in exclude:
                    variables.update(extra_var)
        _LOGGER.debug(
            "Submission template: {}".format(self.compute["submission_template"])
        )
        if output_path:
            _LOGGER.info("Writing script to {}".format(os.path.abspath(output_path)))

        return write_submit_script(output_path, self.template(), variables)

    def _handle_missing_env_attrs(self, config_file, when_missing):
        """Default environment settings aren't required; warn, though."""
        missing_env_attrs = [
            attr
            for attr in [NEW_COMPUTE_KEY, "config_file"]
            if getattr(self, attr, None) is None
        ]
        if not missing_env_attrs:
            return
        message = "'{}' lacks environment attributes: {}".format(
            config_file, missing_env_attrs
        )
        if when_missing is None:
            _LOGGER.warning(message)
        else:
            when_missing(message)


def select_divvy_config(filepath):
    """
    Selects the divvy config file path to load.

    This uses a priority ordering to first choose a config file path if
    it's given, but if not, then look in a priority list of environment
    variables and choose the first available file path to return. If none of
    these options succeed, the default config path will be returned.

    :param str | NoneType filepath: direct file path specification
    :return str: path to the config file to read
    """
    divcfg = yacman.select_config(
        config_filepath=filepath,
        config_env_vars=COMPUTE_SETTINGS_VARNAME,
        default_config_filepath=DEFAULT_CONFIG_FILEPATH,
        check_exist=True,
        config_name="divvy",
    )
    _LOGGER.debug("Selected divvy config: {}".format(divcfg))
    return divcfg


def divvy_init(config_path, template_config_path):
    """
    Initialize a genome config file.

    :param str config_path: path to divvy configuration file to
        create/initialize
    :param str template_config_path: path to divvy configuration file to
        copy FROM
    """
    if not config_path:
        _LOGGER.error("You must specify a file path to initialize.")
        return

    if not template_config_path:
        _LOGGER.error("You must specify a template config file path.")
        return

    if config_path and not os.path.exists(config_path):
        # dcc.write(config_path)
        # Init should *also* write the templates.
        dest_folder = os.path.dirname(config_path)
        copytree(os.path.dirname(template_config_path), dest_folder)
        template_subfolder = os.path.join(dest_folder, "divvy_templates")
        _LOGGER.info("Wrote divvy templates to folder: {}".format(template_subfolder))
        new_template = os.path.join(
            os.path.dirname(config_path), os.path.basename(template_config_path)
        )
        os.rename(new_template, config_path)
        _LOGGER.info("Wrote new divvy configuration file: {}".format(config_path))
    else:
        _LOGGER.warning("Can't initialize, file exists: {} ".format(config_path))



import logging
import os
from .const import *
from .conductor import _get_yaml_path

_LOGGER = logging.getLogger(__name__)


def write_sample_yaml_prj(namespaces):
    """
    Plugin: saves sample representation with project reference to YAML.

    This plugin can be parametrized by providing the path value/template in
    'pipeline.var_templates.sample_yaml_prj_path'. This needs to be a complete and
    absolute path to the file where sample YAML representation is to be
    stored.

    :param dict namespaces: variable namespaces dict
    :return dict: sample namespace dict
    """
    sample = namespaces["sample"]
    sample.to_yaml(
        _get_yaml_path(namespaces, SAMPLE_YAML_PRJ_PATH_KEY, "_sample_prj"),
        add_prj_ref=True,
    )
    return {"sample": sample}


def write_custom_template(namespaces):
    """
    Plugin: Populates a user-provided jinja template

    Parameterize by providing pipeline.var_templates.custom_template
    """

    def load_template(pipeline):
        with open(namespaces["pipeline"]["var_templates"]["custom_template"], "r") as f:
            x = f.read()
        t = jinja2.Template(x)
        return t

    err_msg = (
        "Custom template plugin requires a template in var_templates.custom_template"
    )
    if "var_templates" not in namespaces["pipeline"].keys():
        _LOGGER.error(err_msg)
        return None

    if "custom_template" not in namespaces["pipeline"]["var_templates"].keys():
        _LOGGER.error(err_msg)
        return None

    import jinja2

    tpl = load_template(namespaces["pipeline"])
    content = tpl.render(namespaces)
    pth = _get_yaml_path(namespaces, "custom_template_output", "config")
    namespaces["sample"]["custom_template_output"] = pth
    with open(pth, "wb") as fh:
        # print(content)
        fh.write(content.encode())

    return {"sample": namespaces["sample"]}


def write_sample_yaml_cwl(namespaces):
    """
    Plugin: Produce a cwl-compatible yaml representation of the sample

    Also adds the 'cwl_yaml' attribute to sample objects, which points
    to the file produced.

    This plugin can be parametrized by providing the path value/template in
    'pipeline.var_templates.sample_cwl_yaml_path'. This needs to be a complete and
    absolute path to the file where sample YAML representation is to be
    stored.

    :param dict namespaces: variable namespaces dict
    :return dict: updated variable namespaces dict
    """
    from eido import read_schema
    from ubiquerg import is_url

    def _get_schema_source(
        schema_source, piface_dir=namespaces["looper"]["piface_dir"]
    ):
        # Stolen from piface object; should be a better way to do this...
        if is_url(schema_source):
            return schema_source
        elif not os.path.isabs(schema_source):
            schema_source = os.path.join(piface_dir, schema_source)
        return schema_source

    # To be compatible as a CWL job input, we need to handle the
    # File and Directory object types directly.
    sample = namespaces["sample"]
    sample.sample_yaml_cwl = _get_yaml_path(
        namespaces, SAMPLE_CWL_YAML_PATH_KEY, "_sample_cwl"
    )

    if "input_schema" in namespaces["pipeline"]:
        schema_path = _get_schema_source(namespaces["pipeline"]["input_schema"])
        file_list = []
        for ischema in read_schema(schema_path):
            if "files" in ischema["properties"]["samples"]["items"]:
                file_list.extend(ischema["properties"]["samples"]["items"]["files"])

        for file_attr in file_list:
            _LOGGER.debug("CWL-ing file attribute: {}".format(file_attr))
            file_attr_value = sample[file_attr]
            # file paths are assumed relative to the sample table;
            # but CWL assumes they are relative to the yaml output file,
            # so we convert here.
            file_attr_rel = os.path.relpath(
                file_attr_value, os.path.dirname(sample.sample_yaml_cwl)
            )
            sample[file_attr] = {"class": "File", "path": file_attr_rel}

        directory_list = []
        for ischema in read_schema(schema_path):
            if "directories" in ischema["properties"]["samples"]["items"]:
                directory_list.extend(
                    ischema["properties"]["samples"]["items"]["directories"]
                )

        for dir_attr in directory_list:
            _LOGGER.debug("CWL-ing directory attribute: {}".format(dir_attr))
            dir_attr_value = sample[dir_attr]
            # file paths are assumed relative to the sample table;
            # but CWL assumes they are relative to the yaml output file,
            # so we convert here.
            sample[dir_attr] = {"class": "Directory", "location": dir_attr_value}
    else:
        _LOGGER.warning(
            "No 'input_schema' defined, producing a regular "
            "sample YAML representation"
        )
    _LOGGER.info("Writing sample yaml to {}".format(sample.sample_yaml_cwl))
    sample.to_yaml(sample.sample_yaml_cwl)
    return {"sample": sample}


def write_sample_yaml(namespaces):
    """
    Plugin: saves sample representation to YAML.

    This plugin can be parametrized by providing the path value/template in
    'pipeline.var_templates.sample_yaml_path'. This needs to be a complete and
    absolute path to the file where sample YAML representation is to be
    stored.

    :param dict namespaces: variable namespaces dict
    :return dict: sample namespace dict
    """
    sample = namespaces["sample"]
    sample["sample_yaml_path"] = _get_yaml_path(
        namespaces, SAMPLE_YAML_PATH_KEY, "_sample"
    )
    sample.to_yaml(sample["sample_yaml_path"], add_prj_ref=False)
    return {"sample": sample}



""" Pipeline job submission orchestration """

import importlib
import logging
import os
import subprocess
import time
import yaml
from copy import copy, deepcopy
from json import loads
from subprocess import check_output
from typing import *

from eido import read_schema, get_input_files_size
from eido.const import INPUT_FILE_SIZE_KEY, MISSING_KEY
from jinja2.exceptions import UndefinedError

from peppy.const import CONFIG_KEY, SAMPLE_NAME_ATTR, SAMPLE_YAML_EXT
from peppy.exceptions import RemoteYAMLError
from pipestat import PipestatError
from ubiquerg import expandpath, is_command_callable
from yaml import dump
from yacman import YAMLConfigManager

from .const import *
from .exceptions import JobSubmissionException, SampleFailedException
from .processed_project import populate_sample_paths
from .utils import fetch_sample_flags, jinja_render_template_strictly


_LOGGER = logging.getLogger(__name__)


def _get_yaml_path(namespaces, template_key, default_name_appendix="", filename=None):
    """
    Get a path to a YAML file for the sample.

    :param dict[dict]] namespaces: namespaces mapping
    :param str template_key: the name of the key in 'var_templates' piface
        section that points to a template to render to get the
        user-provided target YAML path
    :param str default_name_appendix: a string to append to insert in target
        YAML file name: '{sample.sample_name}<>.yaml'
    :param str filename: A filename without folders. If not provided, a
        default name of sample_name.yaml will be used.
    :return str: sample YAML file path
    """
    if (
        VAR_TEMPL_KEY in namespaces["pipeline"]
        and template_key in namespaces["pipeline"][VAR_TEMPL_KEY]
    ):
        _LOGGER.debug(f"Sample namespace: {namespaces['sample']}")
        x = jinja_render_template_strictly("{sample.sample_name}", namespaces)
        _LOGGER.debug(f"x: {x}")
        cpy = namespaces["pipeline"][VAR_TEMPL_KEY][template_key]
        _LOGGER.debug(f"cpy: {cpy}")
        path = expandpath(jinja_render_template_strictly(cpy, namespaces))
        _LOGGER.debug(f"path: {path}")

        if not path.endswith(SAMPLE_YAML_EXT) and not filename:
            raise ValueError(
                f"{template_key} is not a valid target YAML file path. "
                f"It needs to end with: {' or '.join(SAMPLE_YAML_EXT)}"
            )
        final_path = os.path.join(path, filename) if filename else path
        if not os.path.exists(os.path.dirname(final_path)):
            os.makedirs(os.path.dirname(final_path), exist_ok=True)
    else:
        # default YAML location
        f = (
            filename
            or f"{namespaces['sample'][SAMPLE_NAME_ATTR]}"
            f"{default_name_appendix}"
            f"{SAMPLE_YAML_EXT[0]}"
        )
        default = os.path.join(namespaces["looper"][OUTDIR_KEY], "submission")
        final_path = os.path.join(default, f)
        if not os.path.exists(default):
            os.makedirs(default, exist_ok=True)

    _LOGGER.debug(f"Writing sample yaml: {final_path}")
    return final_path


def write_pipestat_config(looper_pipestat_config_path, pipestat_config_dict):
    """
    This is run at the project level, not at the sample level.
    """
    with open(looper_pipestat_config_path, "w") as f:
        yaml.dump(pipestat_config_dict, f)
    print(f"Initialized looper config file: {looper_pipestat_config_path}")

    return True


def write_submission_yaml(namespaces):
    """
    Save all namespaces to YAML.

    :param dict namespaces: variable namespaces dict
    :return dict: sample namespace dict
    """
    path = _get_yaml_path(namespaces, SAMPLE_CWL_YAML_PATH_KEY, "_submission")
    my_namespaces = {}
    for namespace, values in namespaces.items():
        my_namespaces.update({str(namespace): dict(values)})
    with open(path, "w") as yamlfile:
        dump(my_namespaces, yamlfile)
    return my_namespaces


class SubmissionConductor(object):
    """
    Collects and then submits pipeline jobs.

    This class holds a 'pool' of commands to submit as a single cluster job.
    Eager to submit a job, each instance's collection of commands expands until
    it reaches the 'pool' has been filled, and it's therefore time to submit the
    job. The pool fills as soon as a fill criteria has been reached, which can
    be either total input file size or the number of individual commands.

    """

    def __init__(
        self,
        pipeline_interface,
        prj,
        delay=0,
        extra_args=None,
        extra_args_override=None,
        ignore_flags=False,
        compute_variables=None,
        max_cmds=None,
        max_size=None,
        automatic=True,
        collate=False,
    ):
        """
        Create a job submission manager.

        The most critical inputs are the pipeline interface and the pipeline
        key, which together determine which provide critical pipeline
        information like resource allocation packages and which pipeline will
        be overseen by this instance, respectively.

        :param PipelineInterface pipeline_interface: Collection of important
            data for one or more pipelines, like resource allocation packages
            and option/argument specifications
        :param prj: Project with which each sample being considered is
            associated (what generated each sample)
        :param float delay: Time (in seconds) to wait before submitting a job
            once it's ready
        :param str extra_args: string to pass to each job generated,
            for example additional pipeline arguments
        :param str extra_args_override: string to pass to each job generated,
            for example additional pipeline arguments. This deactivates the
            'extra' functionality that appends strings defined in
            Sample.command_extra and Project.looper.command_extra to the
            command template.
        :param bool ignore_flags: Whether to ignore flag files present in
            the sample folder for each sample considered for submission
        :param dict[str] compute_variables: A dict with variables that will be made
            available to the compute package. For example, this should include
            the name of the cluster partition to which job or jobs will be submitted
        :param int | NoneType max_cmds: Upper bound on number of commands to
            include in a single job script.
        :param int | float | NoneType max_size: Upper bound on total file
            size of inputs used by the commands lumped into single job script.
        :param bool automatic: Whether the submission should be automatic once
            the pool reaches capacity.
        :param bool collate: Whether a collate job is to be submitted (runs on
            the project level, rather that on the sample level)
        """
        super(SubmissionConductor, self).__init__()
        self.collate = collate
        self.section_key = PROJECT_PL_KEY if self.collate else SAMPLE_PL_KEY
        self.pl_iface = pipeline_interface
        self.pl_name = self.pl_iface.pipeline_name
        self.prj = prj
        self.compute_variables = compute_variables
        self.extra_pipe_args = extra_args
        self.override_extra = False
        if extra_args_override:
            self.extra_pipe_args = extra_args_override
            self.override_extra = True
        self.ignore_flags = ignore_flags

        self.dry_run = self.prj.dry_run
        self.delay = float(delay)
        self._num_good_job_submissions = 0
        self._num_total_job_submissions = 0
        self._num_cmds_submitted = 0
        self._curr_size = 0
        self._failed_sample_names = []
        self._curr_skip_pool = []

        if self.extra_pipe_args:
            _LOGGER.debug(
                "String appended to every pipeline command: "
                "{}".format(self.extra_pipe_args)
            )

        if not self.collate:
            self.automatic = automatic
            if max_cmds is None and max_size is None:
                self.max_cmds = 1
            elif (max_cmds is not None and max_cmds < 1) or (
                max_size is not None and max_size < 0
            ):
                raise ValueError(
                    "If specified, max per-job command count must positive, "
                    "and max per-job total file size must be nonnegative"
                )
            else:
                self.max_cmds = max_cmds
            self.max_size = max_size or float("inf")

            self._pool = []
            self._reset_curr_skips()
            self._skipped_sample_pools = []

    @property
    def failed_samples(self):
        return self._failed_sample_names

    @property
    def num_cmd_submissions(self):
        """
        Return the number of commands that this conductor has submitted.

        :return int: Number of commands submitted so far.
        """
        return self._num_cmds_submitted

    @property
    def num_job_submissions(self):
        """
        Return the number of jobs that this conductor has submitted.

        :return int: Number of jobs submitted so far.
        """
        return self._num_good_job_submissions

    def is_project_submittable(self, force=False):
        """
        Check whether the current project has been already submitted

        :param bool frorce: whether to force the project submission (ignore status/flags)
        """
        if self.prj.pipestat_configured_project:
            psm = self.prj.get_pipestat_managers(project_level=True)[self.pl_name]
            status = psm.get_status()
            if not force and status is not None:
                _LOGGER.info(f"> Skipping project. Determined status: {status}")
                return False
        return True

    def add_sample(self, sample, rerun=False):
        """
        Add a sample for submission to this conductor.

        :param peppy.Sample sample: sample to be included with this conductor's
            currently growing collection of command submissions
        :param bool rerun: whether the given sample is being rerun rather than
            run for the first time
        :return bool: Indication of whether the given sample was added to
            the current 'pool.'
        :raise TypeError: If sample subtype is provided but does not extend
            the base Sample class, raise a TypeError.
        """
        _LOGGER.debug(
            "Adding {} to conductor for {} to {}run".format(
                sample.sample_name, self.pl_name, "re" if rerun else ""
            )
        )
        if self.prj.pipestat_configured:
            psms = self.prj.get_pipestat_managers(sample_name=sample.sample_name)
            sample_statuses = psms[self.pl_name].get_status(
                record_identifier=sample.sample_name
            )
            if sample_statuses == "failed" and rerun is True:
                psms[self.pl_name].set_status(
                    record_identifier=sample.sample_name, status_identifier="waiting"
                )
                sample_statuses = "waiting"
            sample_statuses = [sample_statuses] if sample_statuses else []
        else:
            sample_statuses = fetch_sample_flags(self.prj, sample, self.pl_name)

        use_this_sample = True  # default to running this sample
        msg = None
        if sample_statuses:
            status_str = ", ".join(sample_statuses)
            failed_flag = any("failed" in x for x in sample_statuses)
            if self.ignore_flags:
                msg = f"> Found existing status: {status_str}. Ignoring."
            else:  # this pipeline already has a status
                msg = f"> Found existing status: {status_str}. Skipping sample."
                if failed_flag:
                    msg += " Use rerun to ignore failed status."  # help guidance
                use_this_sample = False
            if rerun:
                # Rescue the sample if rerun requested, and failed flag is found
                if failed_flag:
                    msg = f"> Re-running failed sample. Status: {status_str}"
                    use_this_sample = True
                else:
                    msg = f"> Skipping sample because rerun requested, but no failed flag found. Status: {status_str}"
                    use_this_sample = False
        if msg:
            _LOGGER.info(msg)

        skip_reasons = []
        validation = {}
        validation.setdefault(INPUT_FILE_SIZE_KEY, 0)
        # Check for any missing requirements before submitting.
        _LOGGER.debug("Determining missing requirements")
        schema_source = self.pl_iface.get_pipeline_schemas()
        if schema_source and self.prj.file_checks:
            try:
                validation = get_input_files_size(sample, read_schema(schema_source))
            except RemoteYAMLError:
                _LOGGER.warn(
                    "Could not read remote schema. Skipping inputs validation."
                )
            else:
                if validation[MISSING_KEY]:
                    missing_reqs_msg = (
                        f"Missing files: {', '.join(validation[MISSING_KEY])}"
                    )
                    _LOGGER.warning(NOT_SUB_MSG.format(missing_reqs_msg))
                    use_this_sample and skip_reasons.append("Missing files")

        if _use_sample(use_this_sample, skip_reasons):
            self._pool.append(sample)
            self._curr_size += float(validation[INPUT_FILE_SIZE_KEY])
            if self.automatic and self._is_full(self._pool, self._curr_size):
                self.submit()
        else:
            self._curr_skip_size += float(validation[INPUT_FILE_SIZE_KEY])
            self._curr_skip_pool.append(sample)
            self.write_script(self._curr_skip_pool, self._curr_skip_size)
            self._reset_curr_skips()

        return skip_reasons

    def submit(self, force=False):
        """
        Submit one or more commands as a job.

        This call will submit the commands corresponding to the current pool
        of samples if and only if the argument to 'force' evaluates to a
        true value, or the pool of samples is full.

        :param bool force: Whether submission should be done/simulated even
            if this conductor's pool isn't full.
        :return bool: Whether a job was submitted (or would've been if
            not for dry run)
        """
        submitted = False
        if not self._pool:
            _LOGGER.debug("No submission (no pooled samples): %s", self.pl_name)
            # submitted = False
        elif self.collate or force or self._is_full(self._pool, self._curr_size):
            if not self.collate:
                for s in self._pool:
                    schemas = self.prj.get_schemas(
                        self.prj.get_sample_piface(s[self.prj.sample_table_index]),
                        OUTPUT_SCHEMA_KEY,
                    )

                    for schema in schemas:
                        populate_sample_paths(s, read_schema(schema)[0])

            script = self.write_script(self._pool, self._curr_size)
            # Determine whether to actually do the submission.
            _LOGGER.info(
                "Job script (n={0}; {1:.2f}Gb): {2}".format(
                    len(self._pool), self._curr_size, script
                )
            )
            if self.dry_run:
                _LOGGER.info("Dry run, not submitted")
            elif self._rendered_ok:
                sub_cmd = self.prj.dcc.compute["submission_command"]
                submission_command = "{} {}".format(sub_cmd, script)
                # Capture submission command return value so that we can
                # intercept and report basic submission failures; #167
                try:
                    subprocess.check_call(submission_command, shell=True)
                except subprocess.CalledProcessError:
                    fails = (
                        "" if self.collate else [s.sample_name for s in self._samples]
                    )
                    self._failed_sample_names.extend(fails)
                    self._reset_pool()
                    raise JobSubmissionException(sub_cmd, script)
                time.sleep(self.delay)

            # Update the job and command submission tallies.
            _LOGGER.debug("SUBMITTED")
            if self._rendered_ok:
                submitted = True
                self._num_cmds_submitted += len(self._pool)
            self._reset_pool()

        else:
            _LOGGER.debug(
                f"No submission (pool is not full and submission was not forced): {self.pl_name}"
            )
            # submitted = False

        return submitted

    def _is_full(self, pool, size):
        """
        Determine whether it's time to submit a job for the pool of commands.

        Instances of this class maintain a sort of 'pool' of commands that
        expands as each new command is added, until a time that it's deemed
        'full' and th

        :return bool: Whether this conductor's pool of commands is 'full' and
            ready for submission, as determined by its parameterization
        """
        return self.max_cmds == len(pool) or size >= self.max_size

    @property
    def _samples(self):
        """
        Return a collection of pooled samples.

        :return Iterable[str]: collection of samples currently in the active
            pool for this submission conductor
        """
        return [s for s in self._pool]

    def _sample_lump_name(self, pool):
        """Determine how to refer to the 'sample' for this submission."""
        if self.collate:
            return self.prj.name
        if 1 == self.max_cmds:
            assert 1 == len(pool), (
                "If there's a single-command limit on job submission, jobname"
                " must be determined with exactly one sample in the pool,"
                " but there is/are {}.".format(len(pool))
            )
            sample = pool[0]
            return sample.sample_name
        else:
            # Note the order in which the increment of submission count and
            # the call to this function can influence naming. Make the jobname
            # generation call (this method) before incrementing the
            # submission counter, but add 1 to the index so that we get a
            # name concordant with 1-based, not 0-based indexing.
            return "lump{}".format(self._num_total_job_submissions + 1)

    def _jobname(self, pool):
        """Create the name for a job submission."""
        return "{}_{}".format(self.pl_iface.pipeline_name, self._sample_lump_name(pool))

    def _build_looper_namespace(self, pool, size):
        """
        Compile a mapping of looper/submission related settings for use in
        the command templates and in submission script creation
        in divvy (via adapters).

        :param Iterable[peppy.Sample] pool: collection of sample instances
        :param float size: cumulative size of the given pool
        :return yacman.YAMLConfigManager: looper/submission related settings
        """
        settings = YAMLConfigManager()
        settings["config_file"] = self.prj.config_file
        settings["pep_config"] = self.prj.pep_config

        settings[RESULTS_SUBDIR_KEY] = self.prj.results_folder
        settings[SUBMISSION_SUBDIR_KEY] = self.prj.submission_folder
        settings[OUTDIR_KEY] = self.prj.output_dir
        # TODO: explore pulling additional constants from peppy and/or divvy.
        settings["sample_output_folder"] = os.path.join(
            self.prj.results_folder, self._sample_lump_name(pool)
        )
        settings[JOB_NAME_KEY] = self._jobname(pool)
        settings["total_input_size"] = size
        settings["log_file"] = (
            os.path.join(self.prj.submission_folder, settings[JOB_NAME_KEY]) + ".log"
        )
        settings["piface_dir"] = os.path.dirname(self.pl_iface.pipe_iface_file)
        if hasattr(self.prj, "pipeline_config"):
            # Make sure it's a file (it could be provided as null.)
            pl_config_file = self.prj.pipeline_config
            if pl_config_file:
                if not os.path.isfile(pl_config_file):
                    _LOGGER.error(
                        "Pipeline config file specified " "but not found: %s",
                        pl_config_file,
                    )
                    raise IOError(pl_config_file)
                _LOGGER.info("Found config file: %s", pl_config_file)
                # Append arg for config file if found
                settings["pipeline_config"] = pl_config_file
        return settings

    def _set_pipestat_namespace(
        self, sample_name: Optional[str] = None
    ) -> YAMLConfigManager:
        """
        Compile a mapping of pipestat-related settings for use in
        the command templates. Accessible via: {pipestat.attrname}

        :param str sample_name: name of the sample to get the pipestat
            namespace for. If not provided the pipestat namespace will
            be determined based on the Project
        :return yacman.YAMLConfigManager: pipestat namespace
        """
        try:
            psms = (
                self.prj.get_pipestat_managers(sample_name)
                if sample_name
                else self.prj.get_pipestat_managers(project_level=True)
            )
            psm = psms[self.pl_iface.pipeline_name]
        except (PipestatError, AttributeError) as e:
            # pipestat section faulty or not found in project.looper or sample
            # or project is missing required pipestat attributes
            _LOGGER.debug(
                f"Could not determine pipestat namespace. Caught exception: "
                f"{getattr(e, 'message', repr(e))}"
            )
            # return an empty mapping
            return YAMLConfigManager()
        else:
            full_namespace = {
                "results_file": psm.file,
                "record_identifier": psm.record_identifier,
                "config_file": psm.config_path,
            }
            filtered_namespace = {k: v for k, v in full_namespace.items() if v}
            return YAMLConfigManager(filtered_namespace)

    def write_script(self, pool, size):
        """
        Create the script for job submission.

        :param Iterable[peppy.Sample] pool: collection of sample instances
        :param float size: cumulative size of the given pool
        :return str: Path to the job submission script created.
        """
        # looper settings determination
        if self.collate:
            pool = [None]
        looper = self._build_looper_namespace(pool, size)
        commands = []
        namespaces = dict(
            project=self.prj[CONFIG_KEY],
            looper=looper,
            pipeline=self.pl_iface,
            compute=self.prj.dcc.compute,
        )
        templ = self.pl_iface["command_template"]
        if not self.override_extra:
            extras_template = (
                EXTRA_PROJECT_CMD_TEMPLATE
                if self.collate
                else EXTRA_SAMPLE_CMD_TEMPLATE
            )
            templ += extras_template
        for sample in pool:
            # cascading compute settings determination:
            # divcfg < pipeline interface < config <  CLI
            cli = self.compute_variables or {}  # CLI
            if sample:
                namespaces.update({"sample": sample})
            else:
                namespaces.update({"samples": self.prj.samples})
            if self.prj.pipestat_configured:
                pipestat_namespace = self._set_pipestat_namespace(
                    sample_name=sample.sample_name if sample else None
                )
                namespaces.update({"pipestat": pipestat_namespace})
            else:
                # Pipestat isn't configured, simply place empty YAMLConfigManager object instead.
                pipestat_namespace = YAMLConfigManager()
                namespaces.update({"pipestat": pipestat_namespace})
            res_pkg = self.pl_iface.choose_resource_package(
                namespaces, size or 0
            )  # config
            res_pkg.update(cli)
            self.prj.dcc.compute.update(res_pkg)  # divcfg
            namespaces["compute"].update(res_pkg)
            # Here we make a copy of this so that each iteration gets its own template values
            pl_iface = {}
            pl_iface.update(self.pl_iface)
            pl_iface[VAR_TEMPL_KEY] = self.pl_iface.render_var_templates(
                namespaces=namespaces
            )
            _LOGGER.debug(f"namespace pipelines: { pl_iface }")

            namespaces["pipeline"]["var_templates"] = pl_iface[VAR_TEMPL_KEY] or {}
            for k, v in namespaces["pipeline"]["var_templates"].items():
                namespaces["pipeline"]["var_templates"][k] = expandpath(v)

            # pre_submit hook namespace updates
            namespaces = _exec_pre_submit(pl_iface, namespaces)
            self._rendered_ok = False
            try:
                argstring = jinja_render_template_strictly(
                    template=templ, namespaces=namespaces
                )
                print(argstring)
            except UndefinedError as jinja_exception:
                _LOGGER.warning(NOT_SUB_MSG.format(str(jinja_exception)))
            except KeyError as e:
                exc = "pipeline interface is missing {} section".format(str(e))
                _LOGGER.warning(NOT_SUB_MSG.format(exc))
            else:
                commands.append("{} {}".format(argstring, self.extra_pipe_args))
                self._rendered_ok = True
                if sample not in self._curr_skip_pool:
                    self._num_good_job_submissions += 1
                    self._num_total_job_submissions += 1

        looper["command"] = "\n".join(commands)
        if self.collate:
            _LOGGER.debug("samples namespace:\n{}".format(self.prj.samples))
        else:
            _LOGGER.debug(
                "sample namespace:\n{}".format(
                    sample.__str__(max_attr=len(list(sample.keys())))
                )
            )
        _LOGGER.debug("project namespace:\n{}".format(self.prj[CONFIG_KEY]))
        _LOGGER.debug("pipeline namespace:\n{}".format(self.pl_iface))
        _LOGGER.debug("compute namespace:\n{}".format(self.prj.dcc.compute))
        _LOGGER.debug("looper namespace:\n{}".format(looper))
        _LOGGER.debug("pipestat namespace:\n{}".format(pipestat_namespace))
        subm_base = os.path.join(
            expandpath(self.prj.submission_folder), looper[JOB_NAME_KEY]
        )
        return self.prj.dcc.write_script(
            output_path=subm_base + ".sub", extra_vars=[{"looper": looper}]
        )

    def _reset_pool(self):
        """Reset the state of the pool of samples"""
        self._pool = []
        self._curr_size = 0

    def _reset_curr_skips(self):
        self._curr_skip_pool = []
        self._curr_skip_size = 0


def _use_sample(flag, skips):
    return flag and not skips


def _exec_pre_submit(piface, namespaces):
    """
    Execute pre submission hooks defined in the pipeline interface

    :param PipelineInterface piface: piface, a source of pre_submit hooks to execute
    :param dict[dict[]] namespaces: namspaces mapping
    :return dict[dict[]]: updated namspaces mapping
    """

    def _update_namespaces(x, y, cmd=False):
        """
        Update namespaces mapping with a dictionary of the same structure,
        that includes just the values that need to be updated.

        :param dict[dict] x: namespaces mapping
        :param dict[dict] y: mapping to update namespaces with
        :param bool cmd: whether the mapping to update with comes from the
            command template, used for messaging
        """
        if not y:
            return
        if not isinstance(y, dict):
            if cmd:
                raise TypeError(
                    f"Object returned by {PRE_SUBMIT_HOOK_KEY}."
                    f"{PRE_SUBMIT_CMD_KEY} must return a dictionary when "
                    f"processed with json.loads(), not {y.__class__.__name__}"
                )
            raise TypeError(
                f"Object returned by {PRE_SUBMIT_HOOK_KEY}."
                f"{PRE_SUBMIT_PY_FUN_KEY} must return a dictionary,"
                f" not {y.__class__.__name__}"
            )
        _LOGGER.debug("Updating namespaces with:\n{}".format(y))
        for namespace, mapping in y.items():
            for key, val in mapping.items():
                x[namespace][key] = val

    if PRE_SUBMIT_HOOK_KEY in piface:
        pre_submit = piface[PRE_SUBMIT_HOOK_KEY]
        if PRE_SUBMIT_PY_FUN_KEY in pre_submit:
            for py_fun in pre_submit[PRE_SUBMIT_PY_FUN_KEY]:
                pkgstr, funcstr = os.path.splitext(py_fun)
                pkg = importlib.import_module(pkgstr)
                func = getattr(pkg, funcstr[1:])
                _LOGGER.info(
                    "Calling pre-submit function: {}.{}".format(pkgstr, func.__name__)
                )
                _update_namespaces(namespaces, func(namespaces))
        if PRE_SUBMIT_CMD_KEY in pre_submit:
            for cmd_template in pre_submit[PRE_SUBMIT_CMD_KEY]:
                _LOGGER.debug(
                    "Rendering pre-submit command template: {}".format(cmd_template)
                )
                try:
                    cmd = jinja_render_template_strictly(
                        template=cmd_template, namespaces=namespaces
                    )
                    _LOGGER.info("Executing pre-submit command: {}".format(cmd))
                    json = loads(check_output(cmd, shell=True))
                except Exception as e:
                    if hasattr(e, "output"):
                        print(e.output)
                    _LOGGER.error(
                        "Could not retrieve JSON via command: '{}'".format(cmd)
                    )
                    raise
                else:
                    _update_namespaces(namespaces, json, cmd=True)
    return namespaces



"""
Processed Project manipulation functions.
These functions are used to process pipestat-compatible schema,
but the report generation approach has changed.
"""

__author__ = "Michal Stolarczyk"
__email__ = "michal@virginia.edu"

# import os
# from collections.abc import Mapping
# from copy import copy
# from logging import getLogger

# from eido.const import *
# from eido.exceptions import *
# from peppy.project import Project
# from peppy.sample import Sample
# from pipestat import SCHEMA_TYPE_KEY, SchemaError
# from ubiquerg import expandpath

# _LOGGER = getLogger(__name__)
# PATH_KEY = "path"
# THUMB_PATH_KEY = "thumbnail_path"
# PATH_LIKE = [PATH_KEY, THUMB_PATH_KEY]


# def _populate_paths_in_schema(object, schema):
#     """
#     Populate path-like object attributes with other object attributes
#     based on a defined template, e.g. '/Users/x/test_{name}/{genome}_file.txt'

#     :param Mapping object: object with attributes to populate path template with
#     :param dict schema: schema with path attributes defined, e.g.
#         output of read_schema function
#     :return Mapping: object with path templates populated
#     """

#     def _recurse_and_populate(mapping, object):
#         """
#         Recursively populate paths and thumbnail_paths templates in a mapping

#         :param any mapping: a potential mapping with paths to populate
#         :param Mapping object: object with attributes to populate path
#             template with
#         :return any: potentially populated object
#         """
#         if isinstance(mapping, Mapping):
#             for k, v in mapping.items():
#                 if isinstance(v, Mapping):
#                     _recurse_and_populate(v, object)
#                 elif k in PATH_LIKE:
#                     try:
#                         mapping[k] = expandpath(v.format(**dict(object.items())))
#                     except Exception as e:
#                         _LOGGER.warning(
#                             f"Caught exception: {getattr(e, 'message', repr(e))}."
#                             f"\nCould not populate template in schema: {v}"
#                         )
#                     else:
#                         _LOGGER.debug(f"Populated: {mapping[k]}")
#         return mapping

#     for k, v in schema.items():
#         if "value" not in v:
#             continue
#         if SCHEMA_TYPE_KEY not in v:
#             raise SchemaError(
#                 f"'{SCHEMA_TYPE_KEY}' not found in '{k}' section "
#                 f"of the output schema"
#             )
#         schema[k] = _recurse_and_populate(v, object)
#     return schema


# def populate_sample_paths(sample, schema):
#     """
#     Populate path-like Sample attributes with other object attributes
#     based on a defined template, e.g. '/Users/x/test_{name}/{genome}_file.txt'

#     :param peppy.Sample sample: sample to populate paths in
#     :param Iterable[dict] schema: schema with path attributes defined, e.g.
#         output of read_schema function
#     :return Mapping: Sample with path templates populated
#     """
#     # TODO: merge this and 'populate_project_paths' into one?
#     if not isinstance(sample, Sample):
#         raise TypeError("Can only populate paths in peppy.Sample objects")
#     for k, v in _populate_paths_in_schema(sample, copy(schema)).items():
#         if "value" in v:
#             setattr(sample, k, v["value"])
#     return sample


# def populate_project_paths(project, schema):
#     """
#     Populate path-like Project attributes with other object attributes
#     based on a defined template, e.g. '/Users/x/test_{name}/{genome}_file.txt'

#     :param peppy.Project project: project to populate paths in
#     :param dict schema: schema with path attributes defined, e.g.
#         output of read_schema function
#     :return Mapping: Project with path templates populated
#     """
#     for k, v in _populate_paths_in_schema(project.config, copy(schema)).items():
#         if "value" in v:
#             setattr(project, k, v["value"])
#     return project


# def get_project_outputs(project, schema):
#     """
#     Get project level outputs, where the path-like attributes are populated with
#     project attributes

#     :param peppy.Project project: project o get the set of outputs for
#     :param Iterable[dict] schema: pipestat schema to source the outputs for
#     :return attmap.PathExAttMap: mapping with populated path-like attributes
#     """
#     from attmap import PathExAttMap

#     schema = schema[-1]  # use only first schema, in case there are imports
#     populated = populate_project_paths(project, schema)
#     return PathExAttMap({k: getattr(populated, k) for k in schema.keys()})
"""
Processed Project manipulation functions, required pipelines with no pipestat support (old schema)
"""
import os
from logging import getLogger

from eido.const import *
from eido.exceptions import *
from peppy.project import Project
from peppy.sample import Sample

_LOGGER = getLogger(__name__)
PATH_KEY = "path"
THUMB_PATH_KEY = "thumbnail_path"
PATH_LIKE = [PATH_KEY, THUMB_PATH_KEY]


def _get_path_sect_keys(mapping, keys=[PATH_KEY]):
    """
    Get names of subsections in a mapping that contain collection of keys

    :param Mapping mapping: schema subsection to search for paths
    :param  Iterable[str] keys: collection of keys to check for
    :return Iterable[str]: collection of keys to path-like sections
    """
    return [k for k, v in mapping.items() if bool(set(keys) & set(mapping[k]))]


def _populate_paths(object, schema, check_exist):
    """
    Populate path-like object attributes with other object attributes
    based on a defined template, e.g. '/Users/x/test_{name}/{genome}_file.txt'

    :param Mapping object: object with attributes to populate path template with
    :param dict schema: schema with path attributes defined, e.g.
        output of read_schema function
    :param bool check_exist: whether the paths should be check for existence
    :return Mapping: object with path templates populated
    """
    if PROP_KEY not in schema:
        raise EidoSchemaInvalidError("Schema is missing properties section.")
    missing = []
    s = schema[PROP_KEY]
    path_sects = _get_path_sect_keys(s)
    for ps in path_sects:
        templ = s[ps][PATH_KEY]
        try:
            populated = templ.format(**dict(object.items()))
        except Exception as e:
            _LOGGER.warning(
                "Caught exception: {}.\n"
                "Could not populate path: {}".format(
                    getattr(e, "message", repr(e)), templ
                )
            )
        else:
            setattr(object, ps, populated)
            _LOGGER.debug("Path set to: {}".format(object[ps]))
            if check_exist and not os.path.exists(object[ps]):
                missing.append(object[ps])
    if missing:
        raise PathAttrNotFoundError(
            "Path attributes not found:\n- {}".format("\n- ".join(missing))
        )


def populate_sample_paths(sample, schema, check_exist=False):
    """
    Populate path-like Sample attributes with other object attributes
    based on a defined template, e.g. '/Users/x/test_{name}/{genome}_file.txt'

    :param peppy.Sample sample: sample to populate paths in
    :param Iterable[dict] schema: schema with path attributes defined, e.g.
        output of read_schema function
    :param bool check_exist: whether the paths should be check for existence
    :return Mapping: Sample with path templates populated
    """
    if not isinstance(sample, Sample):
        raise TypeError("Can only populate paths in peppy.Sample objects")
    # schema = schema[-1]  # use only first schema, in case there are imports
    if PROP_KEY in schema and "samples" in schema[PROP_KEY]:
        _populate_paths(sample, schema, check_exist)


def populate_project_paths(project, schema, check_exist=False):
    """
    Populate path-like Project attributes with other object attributes
    based on a defined template, e.g. '/Users/x/test_{name}/{genome}_file.txt'

    :param peppy.Project project: project to populate paths in
    :param dict schema: schema with path attributes defined, e.g.
        output of read_schema function
    :param bool check_exist: whether the paths should be check for existence
    :return Mapping: Project with path templates populated
    """
    if not isinstance(project, Project):
        raise TypeError("Can only populate paths in peppy.Project objects")
    _populate_paths(project, schema, check_exist)


def get_project_outputs(project, schema):
    """
    Get project level outputs with path-like attributes populated with
    project attributes

    :param peppy.Project project:
    :param Iterable[dict] schema:
    :return yacman.YAMLConfigManager: mapping with populated path-like attributes
    """
    from yacman import YAMLConfigManager

    # if not any([isinstance(project, Project),
    #             issubclass(type(project), Project)]):
    #     raise TypeError("Can only populate paths in peppy.Project "
    #                     "objects or it subclasses")
    schema = schema[-1]  # use only first schema, in case there are imports
    if PROP_KEY not in schema:
        raise EidoSchemaInvalidError("Schema is missing properties section.")
    res = {}
    s = schema[PROP_KEY]
    path_sects = _get_path_sect_keys(s, keys=PATH_LIKE)
    for ps in path_sects:
        res[ps] = s[ps]
        for p in PATH_LIKE:
            try:
                res[ps][p] = s[ps][p].format(**dict(project.items()))
            except Exception as e:
                _LOGGER.debug(
                    "Caught exception: {}.\n Could not populate {} "
                    "path".format(p, str(e))
                )
    return YAMLConfigManager(res)



"""Tests for determination of desired sample range"""

from itertools import chain
import pytest
from hypothesis import given, strategies as st
from looper.utils import (
    NatIntervalException,
    desired_samples_range_limited,
    desired_samples_range_skipped,
)


@pytest.mark.parametrize(
    "arg_and_limit",
    [
        ("2:3", 1),
        ("2-3", 1),
        ("1,3", 2),
        ("1,3", 4),
        ("1;3", 2),
        ("1;3", 4),
        ("1/3", 2),
        ("1/3", 4),
        ("1 3", 2),
        ("1 3", 4),
        ("", 10),
        (":", 10),
        ("-", 10),
        (" - ", 10),
        (" : ", 10),
    ],
)
@pytest.mark.parametrize(
    "func", [desired_samples_range_limited, desired_samples_range_skipped]
)
def test_invalid_input__raise_expected_error(arg_and_limit, func):
    arg, num_samples = arg_and_limit
    with pytest.raises(NatIntervalException):
        func(arg, num_samples)


def assert_iterable_equality(obs, exp):
    assert list(obs) == list(exp)


@pytest.mark.parametrize(
    ["arg", "num_samples", "func", "expected"],
    [
        ("2-6", 4, desired_samples_range_limited, range(2, 5)),
        ("3:6", 8, desired_samples_range_limited, range(3, 7)),
        ("4", 3, desired_samples_range_limited, range(1, 4)),
        ("4", 6, desired_samples_range_limited, range(1, 5)),
        ("2:6", 4, desired_samples_range_skipped, range(1, 2)),
        ("3-6", 8, desired_samples_range_skipped, chain(range(1, 3), range(7, 9))),
        ("4", 3, desired_samples_range_skipped, []),
        ("4", 6, desired_samples_range_skipped, range(5, 7)),
    ],
)
def test_valid_input__yields_correct_range(arg, num_samples, func, expected):
    observed = func(arg, num_samples=num_samples)
    assert_iterable_equality(obs=observed, exp=expected)



#!/usr/bin/env python
"""
Looper: a pipeline submission engine. https://github.com/pepkit/looper
"""

import abc
import argparse
import csv
import logging
import subprocess
import yaml
import os
import pandas as _pd

# Need specific sequence of actions for colorama imports?
from colorama import init

init()
from shutil import rmtree

# from collections.abc import Mapping
from collections import defaultdict
from colorama import Fore, Style
from eido import validate_config, validate_sample
from eido.exceptions import EidoValidationError
from jsonschema import ValidationError
from peppy.const import *
from peppy.exceptions import RemoteYAMLError
from rich.color import Color
from rich.console import Console
from rich.table import Table
from ubiquerg.cli_tools import query_yes_no
from ubiquerg.collection import uniqify


from .conductor import SubmissionConductor

from .exceptions import *
from .const import *
from .pipeline_interface import PipelineInterface
from .project import Project
from .utils import (
    desired_samples_range_skipped,
    desired_samples_range_limited,
    sample_folder,
)
from pipestat.reports import get_file_for_table
from pipestat.reports import get_file_for_project

_PKGNAME = "looper"
_LOGGER = logging.getLogger(_PKGNAME)


class Executor(object):
    """Base class that ensures the program's Sample counter starts.

    Looper is made up of a series of child classes that each extend the base
    Executor class. Each child class does a particular task (such as run the
    project, summarize the project, destroy the project, etc). The parent
    Executor class simply holds the code that is common to all child classes,
    such as counting samples as the class does its thing."""

    __metaclass__ = abc.ABCMeta

    def __init__(self, prj):
        """
        The Project defines the instance; establish an iteration counter.

        :param Project prj: Project with which to work/operate on
        """
        super(Executor, self).__init__()
        self.prj = prj
        self.counter = LooperCounter(len(prj.samples))

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        """Do the work of the subcommand/program."""
        pass


class Checker(Executor):
    def __call__(self, args):
        """
        Check Project status, using pipestat.

        :param argparse.Namespace: arguments provided to the command
        """

        # aggregate pipeline status data
        status = {}
        if args.project:
            psms = self.prj.get_pipestat_managers(project_level=True)
            for pipeline_name, psm in psms.items():
                s = psm.get_status() or "unknown"
                status.setdefault(pipeline_name, {})
                status[pipeline_name][self.prj.name] = s
                _LOGGER.debug(f"{self.prj.name} ({pipeline_name}): {s}")
        else:
            for sample in self.prj.samples:
                psms = self.prj.get_pipestat_managers(sample_name=sample.sample_name)
                for pipeline_name, psm in psms.items():
                    s = psm.get_status(record_identifier=sample.sample_name)
                    status.setdefault(pipeline_name, {})
                    status[pipeline_name][sample.sample_name] = s
                    _LOGGER.debug(f"{sample.sample_name} ({pipeline_name}): {s}")

        console = Console()

        for pipeline_name, pipeline_status in status.items():
            table_title = f"'{pipeline_name}' pipeline status summary"
            table = Table(
                show_header=True,
                header_style="bold magenta",
                title=table_title,
                width=len(table_title) + 10,
            )
            table.add_column(f"Status", justify="center")
            table.add_column("Jobs count/total jobs", justify="center")
            for status_id in psm.status_schema.keys():
                status_list = list(pipeline_status.values())
                if status_id in status_list:
                    status_count = status_list.count(status_id)
                    table.add_row(status_id, f"{status_count}/{len(status_list)}")
            console.print(table)

        if args.itemized:
            for pipeline_name, pipeline_status in status.items():
                table_title = f"Pipeline: '{pipeline_name}'"
                table = Table(
                    show_header=True,
                    header_style="bold magenta",
                    title=table_title,
                    min_width=len(table_title) + 10,
                )
                table.add_column(
                    f"{'Project' if args.project else 'Sample'} name",
                    justify="right",
                    no_wrap=True,
                )
                table.add_column("Status", justify="center")
                for name, status_id in pipeline_status.items():
                    try:
                        color = Color.from_rgb(
                            *psm.status_schema[status_id]["color"]
                        ).name
                    except KeyError:
                        color = "#bcbcbc"
                        status_id = "unknown"
                    table.add_row(name, f"[{color}]{status_id}[/{color}]")
                console.print(table)

        if args.describe_codes:
            table = Table(
                show_header=True,
                header_style="bold magenta",
                title=f"Status codes description",
                width=len(psm.status_schema_source) + 20,
                caption=f"Descriptions source: {psm.status_schema_source}",
            )
            table.add_column("Status code", justify="center")
            table.add_column("Description", justify="left")
            for status, status_obj in psm.status_schema.items():
                if "description" in status_obj:
                    desc = status_obj["description"]
                else:
                    desc = ""
                table.add_row(status, desc)
            console.print(table)
        return status


class Cleaner(Executor):
    """Remove all intermediate files (defined by pypiper clean scripts)."""

    def __call__(self, args, preview_flag=True):
        """
        Execute the file cleaning process.

        :param argparse.Namespace args: command-line options and arguments
        :param bool preview_flag: whether to halt before actually removing files
        """
        self.counter.show(name=self.prj.name, type="project")
        for sample in self.prj.samples:
            _LOGGER.info(self.counter.show(sample.sample_name))
            sample_output_folder = sample_folder(self.prj, sample)
            cleanup_files = glob.glob(
                os.path.join(sample_output_folder, "*_cleanup.sh")
            )
            if not cleanup_files:
                _LOGGER.info("Nothing to clean.")
                continue
            if preview_flag:
                # Preview: Don't actually clean, just show what will be cleaned.
                _LOGGER.info("Files to clean: %s", ", ".join(cleanup_files))
            else:
                for f in cleanup_files:
                    _LOGGER.info(f)
                    subprocess.call(["sh", f])
        if not preview_flag:
            _LOGGER.info("Clean complete.")
            return 0
        if args.dry_run:
            _LOGGER.info("Dry run. No files cleaned.")
            return 0
        if not args.force_yes and not query_yes_no(
            "Are you sure you want to permanently delete all "
            "intermediate pipeline results for this project?"
        ):
            _LOGGER.info("Clean action aborted by user.")
            return 1
        self.counter.reset()
        return self(args, preview_flag=False)


# NOTE: Adding type hint -> Iterable[Any] gives me  TypeError: 'ABCMeta' object is not subscriptable
def select_samples(prj: Project, args: argparse.Namespace):
    """Use CLI limit/skip arguments to select subset of project's samples."""
    # TODO: get proper element type for signature.
    num_samples = len(prj.samples)
    if args.limit is None and args.skip is None:
        index = range(1, num_samples + 1)
    elif args.skip is not None:
        index = desired_samples_range_skipped(args.skip, num_samples)
    elif args.limit is not None:
        index = desired_samples_range_limited(args.limit, num_samples)
    else:
        raise argparse.ArgumentError(
            "Both --limit and --skip are in use, but they should be mutually exclusive."
        )
    return (prj.samples[i - 1] for i in index)


class Destroyer(Executor):
    """Destroyer of files and folders associated with Project's Samples"""

    def __call__(self, args, preview_flag=True):
        """
        Completely remove all output produced by any pipelines.

        :param argparse.Namespace args: command-line options and arguments
        :param bool preview_flag: whether to halt before actually removing files
        """

        _LOGGER.info("Removing results:")

        for sample in select_samples(prj=self.prj, args=args):
            _LOGGER.info(self.counter.show(sample.sample_name))
            sample_output_folder = sample_folder(self.prj, sample)
            if preview_flag:
                # Preview: Don't actually delete, just show files.
                _LOGGER.info(str(sample_output_folder))
            else:
                _remove_or_dry_run(sample_output_folder, args.dry_run)

        _LOGGER.info("Removing summary:")
        use_pipestat = (
            self.prj.pipestat_configured_project
            if args.project
            else self.prj.pipestat_configured
        )
        if use_pipestat:
            destroy_summary(self.prj, args.dry_run, args.project)
        else:
            _LOGGER.warning(
                "Pipestat must be configured to destroy any created summaries."
            )

        if not preview_flag:
            _LOGGER.info("Destroy complete.")
            return 0

        if args.dry_run:
            _LOGGER.info("Dry run. No files destroyed.")
            return 0

        if not args.force_yes and not query_yes_no(
            "Are you sure you want to permanently delete all pipeline "
            "results for this project?"
        ):
            _LOGGER.info("Destroy action aborted by user.")
            return 1

        self.counter.reset()

        # Finally, run the true destroy:
        return self(args, preview_flag=False)


class Collator(Executor):
    """Submitter for project-level pipelines"""

    def __init__(self, prj):
        """
        Initializes an instance

        :param Project prj: Project with which to work/operate on
        """
        super(Executor, self).__init__()
        self.prj = prj

    def __call__(self, args, **compute_kwargs):
        """
        Matches collators by protocols, creates submission scripts
        and submits them

        :param argparse.Namespace args: parsed command-line options and
            arguments, recognized by looper
        """
        jobs = 0
        self.debug = {}
        project_pifaces = self.prj.project_pipeline_interface_sources
        if not project_pifaces:
            raise MisconfigurationException(
                "Looper requires a pointer to at least one project pipeline. "
                "Please refer to the documentation on linking project to a "
                "pipeline: "
                "http://looper.databio.org/en/latest/defining-a-project"
            )
        self.counter = LooperCounter(len(project_pifaces))
        for project_piface in project_pifaces:
            try:
                project_piface_object = PipelineInterface(
                    project_piface, pipeline_type="project"
                )
            except (IOError, ValidationError) as e:
                _LOGGER.warning(
                    "Ignoring invalid pipeline interface source: {}. "
                    "Caught exception: {}".format(
                        project_piface, getattr(e, "message", repr(e))
                    )
                )
                continue
            _LOGGER.info(
                self.counter.show(
                    name=self.prj.name,
                    type="project",
                    pipeline_name=project_piface_object.pipeline_name,
                )
            )
            conductor = SubmissionConductor(
                pipeline_interface=project_piface_object,
                prj=self.prj,
                compute_variables=compute_kwargs,
                delay=args.time_delay,
                extra_args=args.command_extra,
                extra_args_override=args.command_extra_override,
                ignore_flags=args.ignore_flags,
                collate=True,
            )
            if conductor.is_project_submittable(force=args.ignore_flags):
                conductor._pool = [None]
                conductor.submit()
                jobs += conductor.num_job_submissions
        _LOGGER.info("\nLooper finished")
        _LOGGER.info("Jobs submitted: {}".format(jobs))
        self.debug[DEBUG_JOBS] = jobs
        return self.debug


class Runner(Executor):
    """The true submitter of pipelines"""

    def __call__(self, args, rerun=False, **compute_kwargs):
        """
        Do the Sample submission.

        :param argparse.Namespace args: parsed command-line options and
            arguments, recognized by looper
        :param list remaining_args: command-line options and arguments not
            recognized by looper, germane to samples/pipelines
        :param bool rerun: whether the given sample is being rerun rather than
            run for the first time
        """
        self.debug = {}  # initialize empty dict for return values
        max_cmds = sum(list(map(len, self.prj._samples_by_interface.values())))
        self.counter.total = max_cmds
        failures = defaultdict(list)  # Collect problems by sample.
        processed_samples = set()  # Enforce one-time processing.
        comp_vars = compute_kwargs or {}

        # Determine number of samples eligible for processing.
        num_samples = len(self.prj.samples)

        num_commands_possible = 0
        failed_submission_scripts = []

        # config validation (samples excluded) against all schemas defined
        # for every pipeline matched for this project
        for schema_file in self.prj.get_schemas(self.prj.pipeline_interfaces):
            try:
                validate_config(self.prj, schema_file)
            except RemoteYAMLError:
                _LOGGER.warning(
                    "Could not read remote schema, skipping config validation."
                )

        submission_conductors = {}
        for piface in self.prj.pipeline_interfaces:
            conductor = SubmissionConductor(
                pipeline_interface=piface,
                prj=self.prj,
                compute_variables=comp_vars,
                delay=args.time_delay,
                extra_args=args.command_extra,
                extra_args_override=args.command_extra_override,
                ignore_flags=args.ignore_flags,
                max_cmds=args.lumpn,
                max_size=args.lump,
            )
            submission_conductors[piface.pipe_iface_file] = conductor

        _LOGGER.info(f"Pipestat compatible: {self.prj.pipestat_configured_project}")
        self.debug["Pipestat compatible"] = (
            self.prj.pipestat_configured_project or self.prj.pipestat_configured
        )

        for sample in select_samples(prj=self.prj, args=args):
            pl_fails = []
            skip_reasons = []
            sample_pifaces = self.prj.get_sample_piface(
                sample[self.prj.sample_table_index]
            )
            if not sample_pifaces:
                skip_reasons.append("No pipeline interfaces defined")

            if skip_reasons:
                _LOGGER.warning(NOT_SUB_MSG.format(", ".join(skip_reasons)))
                failures[sample.sample_name] = skip_reasons
                continue

            # single sample validation against a single schema
            # (from sample's piface)
            for schema_file in self.prj.get_schemas(sample_pifaces):
                try:
                    validate_sample(self.prj, sample.sample_name, schema_file)
                except EidoValidationError as e:
                    _LOGGER.error(
                        f"Short-circuiting due to validation error!\nSchema file: "
                        f"{schema_file}\nError: {e}\n{list(e.errors_by_type.keys())}"
                    )
                    self.debug[DEBUG_EIDO_VALIDATION] = (
                        f"Short-circuiting due to validation error!\nSchema file: "
                        f"{schema_file}\nError: {e}\n{list(e.errors_by_type.keys())}"
                    )
                    return False
                except RemoteYAMLError:
                    _LOGGER.warning(
                        f"Could not read remote schema, skipping '{sample.sample_name}' "
                        f"sample validation against {schema_file}"
                    )

            processed_samples.add(sample[self.prj.sample_table_index])

            for sample_piface in sample_pifaces:
                _LOGGER.info(
                    self.counter.show(
                        name=sample.sample_name,
                        pipeline_name=sample_piface.pipeline_name,
                    )
                )
                num_commands_possible += 1
                cndtr = submission_conductors[sample_piface.pipe_iface_file]
                try:
                    curr_pl_fails = cndtr.add_sample(sample, rerun=rerun)
                except JobSubmissionException as e:
                    failed_submission_scripts.append(e.script)
                else:
                    pl_fails.extend(curr_pl_fails)
            if pl_fails:
                failures[sample.sample_name].extend(pl_fails)

        job_sub_total = 0
        cmd_sub_total = 0

        for piface, conductor in submission_conductors.items():
            conductor.submit(force=True)
            job_sub_total += conductor.num_job_submissions
            cmd_sub_total += conductor.num_cmd_submissions

        # Report what went down.
        _LOGGER.info("\nLooper finished")
        _LOGGER.info(
            "Samples valid for job generation: {} of {}".format(
                len(processed_samples), num_samples
            )
        )
        _LOGGER.info("Commands submitted: {} of {}".format(cmd_sub_total, max_cmds))
        self.debug[DEBUG_COMMANDS] = "{} of {}".format(cmd_sub_total, max_cmds)
        if args.dry_run:
            job_sub_total_if_real = job_sub_total
            job_sub_total = 0
            _LOGGER.info(
                f"Dry run. No jobs were actually submitted, but {job_sub_total_if_real} would have been."
            )
        _LOGGER.info("Jobs submitted: {}".format(job_sub_total))
        self.debug[DEBUG_JOBS] = job_sub_total

        # Restructure sample/failure data for display.
        samples_by_reason = defaultdict(set)
        # Collect names of failed sample(s) by failure reason.
        for sample, failures in failures.items():
            for f in failures:
                samples_by_reason[f].add(sample)
                self.debug[f] = sample
        # Collect samples by pipeline with submission failure.
        for piface, conductor in submission_conductors.items():
            # Don't add failure key if there are no samples that failed for
            # that reason.
            if conductor.failed_samples:
                fails = set(conductor.failed_samples)
                samples_by_reason[SUBMISSION_FAILURE_MESSAGE] |= fails

        failed_sub_samples = samples_by_reason.get(SUBMISSION_FAILURE_MESSAGE)
        if failed_sub_samples:
            _LOGGER.info(
                "\n{} samples with at least one failed job submission:"
                " {}".format(len(failed_sub_samples), ", ".join(failed_sub_samples))
            )

        # If failure keys are only added when there's at least one sample that
        # failed for that reason, we can display information conditionally,
        # depending on whether there's actually failure(s).
        if samples_by_reason:
            _LOGGER.info(
                "\n{} unique reasons for submission failure: {}".format(
                    len(samples_by_reason), ", ".join(samples_by_reason.keys())
                )
            )
            full_fail_msgs = [
                _create_failure_message(reason, samples)
                for reason, samples in samples_by_reason.items()
            ]
            _LOGGER.info("\nSummary of failures:\n{}".format("\n".join(full_fail_msgs)))

        if failed_sub_samples:
            _LOGGER.debug("Raising SampleFailedException")
            raise SampleFailedException

        return self.debug


class Reporter(Executor):
    """Combine project outputs into a browsable HTML report"""

    def __call__(self, args):
        # initialize the report builder
        p = self.prj
        project_level = args.project

        if project_level:
            psms = self.prj.get_pipestat_managers(project_level=True)
            print(psms)
            for name, psm in psms.items():
                # Summarize will generate the static HTML Report Function
                report_directory = psm.summarize(looper_samples=self.prj.samples)
                print(f"Report directory: {report_directory}")
        else:
            for piface_source_samples in self.prj._samples_by_piface(
                self.prj.piface_key
            ).values():
                # For each piface_key, we have a list of samples, but we only need one sample from the list to
                # call the related pipestat manager object which will pull ALL samples when using psm.summarize
                first_sample_name = list(piface_source_samples)[0]
                psms = self.prj.get_pipestat_managers(
                    sample_name=first_sample_name, project_level=False
                )
                print(psms)
                for name, psm in psms.items():
                    # Summarize will generate the static HTML Report Function
                    report_directory = psm.summarize(looper_samples=self.prj.samples)
                    print(f"Report directory: {report_directory}")


class Linker(Executor):
    """Create symlinks for reported results. Requires pipestat to be configured."""

    def __call__(self, args):
        # initialize the report builder
        p = self.prj
        project_level = args.project
        link_dir = args.output_dir

        if project_level:
            psms = self.prj.get_pipestat_managers(project_level=True)
            for name, psm in psms.items():
                linked_results_path = psm.link(link_dir=link_dir)
                print(f"Linked directory: {linked_results_path}")
        else:
            for piface_source_samples in self.prj._samples_by_piface(
                self.prj.piface_key
            ).values():
                # For each piface_key, we have a list of samples, but we only need one sample from the list to
                # call the related pipestat manager object which will pull ALL samples when using psm.summarize
                first_sample_name = list(piface_source_samples)[0]
                psms = self.prj.get_pipestat_managers(
                    sample_name=first_sample_name, project_level=False
                )
                for name, psm in psms.items():
                    linked_results_path = psm.link(link_dir=link_dir)
                    print(f"Linked directory: {linked_results_path}")


class Tabulator(Executor):
    """Project/Sample statistics and table output generator

    :return list[str|any] results: list containing output file paths of stats and objects
    """

    def __call__(self, args):
        # p = self.prj
        project_level = args.project
        results = []
        if project_level:
            psms = self.prj.get_pipestat_managers(project_level=True)
            for name, psm in psms.items():
                results = psm.table()
        else:
            for piface_source_samples in self.prj._samples_by_piface(
                self.prj.piface_key
            ).values():
                # For each piface_key, we have a list of samples, but we only need one sample from the list to
                # call the related pipestat manager object which will pull ALL samples when using psm.table
                first_sample_name = list(piface_source_samples)[0]
                psms = self.prj.get_pipestat_managers(
                    sample_name=first_sample_name, project_level=False
                )
                for name, psm in psms.items():
                    results = psm.table()
        # Results contains paths to stats and object summaries.
        return results


def _create_failure_message(reason, samples):
    """Explain lack of submission for a single reason, 1 or more samples."""
    return f"{Fore.LIGHTRED_EX + reason + Style.RESET_ALL}: {', '.join(samples)}"


def _remove_or_dry_run(paths, dry_run=False):
    """
    Remove file or directory or just inform what would be removed in
    case of dry run

    :param list|str paths: list of paths to files/dirs to be removed
    :param bool dry_run: logical indicating whether the files should remain
        untouched and message printed
    """
    paths = paths if isinstance(paths, list) else [paths]
    for path in paths:
        if os.path.exists(path):
            if dry_run:
                _LOGGER.info("DRY RUN. I would have removed: " + path)
            else:
                _LOGGER.info("Removing: " + path)
                if os.path.isfile(path):
                    os.remove(path)
                else:
                    rmtree(path, ignore_errors=True)
        else:
            _LOGGER.info(path + " does not exist.")


def destroy_summary(prj, dry_run=False, project_level=False):
    """
    Delete the summary files if not in dry run mode
    This function is for use with pipestat configured projects.
    """

    if project_level:
        psms = prj.get_pipestat_managers(project_level=True)
        for name, psm in psms.items():
            _remove_or_dry_run(
                [
                    get_file_for_project(
                        psm,
                        pipeline_name=psm["_pipeline_name"],
                        directory="reports",
                    ),
                    get_file_for_table(
                        psm,
                        pipeline_name=psm["_pipeline_name"],
                        appendix="stats_summary.tsv",
                    ),
                    get_file_for_table(
                        psm,
                        pipeline_name=psm["_pipeline_name"],
                        appendix="objs_summary.yaml",
                    ),
                    get_file_for_table(
                        psm, pipeline_name=psm["_pipeline_name"], appendix="reports"
                    ),
                ],
                dry_run,
            )
    else:
        for piface_source_samples in prj._samples_by_piface(prj.piface_key).values():
            # For each piface_key, we have a list of samples, but we only need one sample from the list to
            # call the related pipestat manager object which will pull ALL samples when using psm.table
            first_sample_name = list(piface_source_samples)[0]
            psms = prj.get_pipestat_managers(
                sample_name=first_sample_name, project_level=False
            )
            for name, psm in psms.items():
                _remove_or_dry_run(
                    [
                        get_file_for_project(
                            psm,
                            pipeline_name=psm["_pipeline_name"],
                            directory="reports",
                        ),
                        get_file_for_table(
                            psm,
                            pipeline_name=psm["_pipeline_name"],
                            appendix="stats_summary.tsv",
                        ),
                        get_file_for_table(
                            psm,
                            pipeline_name=psm["_pipeline_name"],
                            appendix="objs_summary.yaml",
                        ),
                        get_file_for_table(
                            psm, pipeline_name=psm["_pipeline_name"], appendix="reports"
                        ),
                    ],
                    dry_run,
                )


class LooperCounter(object):
    """
    Count samples as you loop through them, and create text for the
    subcommand logging status messages.

    :param int total: number of jobs to process
    """

    def __init__(self, total):
        self.count = 0
        self.total = total

    def show(self, name, type="sample", pipeline_name=None):
        """
        Display sample counts status for a particular protocol type.

        The counts are running vs. total for the protocol within the Project,
        and as a side-effect of the call, the running count is incremented.

        :param str name: name of the sample
        :param str type: the name of the level of entity being displayed,
            either project or sample
        :param str pipeline_name: name of the pipeline
        :return str: message suitable for logging a status update
        """
        self.count += 1
        return _submission_status_text(
            type=type,
            curr=self.count,
            total=self.total,
            name=name,
            pipeline_name=pipeline_name,
            color=Fore.CYAN,
        )

    def reset(self):
        self.count = 0

    def __str__(self):
        return "LooperCounter of size {}".format(self.total)


def _submission_status_text(
    curr, total, name, pipeline_name=None, type="sample", color=Fore.CYAN
):
    """Generate submission sample text for run or collate"""
    txt = color + f"## [{curr} of {total}] {type}: {name}"
    if pipeline_name:
        txt += f"; pipeline: {pipeline_name}"
    return txt + Style.RESET_ALL



"""Tests for the natural numbers range data type"""

from typing import *
import pytest
from hypothesis import Phase, given, settings, strategies as st
from looper.utils import NatIntervalException, NatIntervalInclusive


gen_pos_int = st.integers(min_value=1)
gen_opt_int = st.one_of(st.integers(), st.none())


def is_non_pos(opt_int: Optional[int]) -> bool:
    """Determine whether the given value is non-positive (and non-null)."""
    return opt_int is not None and opt_int < 1


def pytest_generate_tests(metafunc):
    if "legit_delim" in metafunc.fixturenames:
        metafunc.parametrize("legit_delim", [":", "-"])


def nondecreasing_pair_strategy(**kwargs):
    """Generate a pair of values in which first respects given upper bound and second is no more than first."""
    return st.tuples(st.integers(**kwargs), st.integers(**kwargs)).filter(
        lambda p: p[0] <= p[1]
    )


class NaturalRangePureConstructorTests:
    """Tests for direct use of natural range primary constructor"""

    @given(upper_bound=gen_pos_int)
    def test_zero_is_prohibited(self, upper_bound):
        """Separate this case since it's an edge case."""
        with pytest.raises(NatIntervalException):
            NatIntervalInclusive(0, upper_bound)

    @given(bounds=nondecreasing_pair_strategy(max_value=0))
    def test_non_positive_is_prohibited(self, bounds):
        lo, hi = bounds
        with pytest.raises(NatIntervalException):
            NatIntervalInclusive(lo, hi)

    @given(bounds=st.tuples(st.integers(), st.integers()).filter(lambda p: p[0] > p[1]))
    def test_upper_less_than_lower__fails_as_expected(self, bounds):
        lo, hi = bounds
        with pytest.raises(NatIntervalException):
            NatIntervalInclusive(lo, hi)


@pytest.mark.skip(reason="Unable to reproduce test failing locally.")
class NaturalRangeFromStringTests:
    """Tests for parsing of natural number range from text, like CLI arg"""

    @pytest.mark.parametrize(
        "arg_template", ["0{sep}0", "{sep}0", "0{sep}", "0{sep}0", "{sep}0", "0{sep}"]
    )
    @given(upper_bound=gen_pos_int)
    def test_zero__does_not_parse(self, arg_template, legit_delim, upper_bound):
        arg = arg_template.format(sep=legit_delim)
        with pytest.raises(NatIntervalException):
            NatIntervalInclusive.from_string(arg, upper_bound=upper_bound)

    @given(upper_bound=st.integers())
    def test_just_delimiter__does_not_parse(self, legit_delim, upper_bound):
        with pytest.raises(NatIntervalException):
            NatIntervalInclusive.from_string(legit_delim, upper_bound=upper_bound)

    @given(
        lo_hi_upper=st.tuples(gen_opt_int, gen_opt_int, st.integers()).filter(
            lambda t: (t[0] is not None or t[1] is not None)
            and any(is_non_pos(n) for n in t)
        )
    )
    def test_nonpositive_values__fail_with_expected_error(
        self, lo_hi_upper, legit_delim
    ):
        lo, hi, upper_bound = lo_hi_upper
        if lo is None and hi is None:
            raise ValueError("Both lower and upper bound generated are null.")
        if lo is None:
            arg = legit_delim + str(hi)
        elif hi is None:
            arg = str(lo) + legit_delim
        else:
            arg = str(lo) + legit_delim + str(hi)
        with pytest.raises(NatIntervalException):
            NatIntervalInclusive.from_string(arg, upper_bound=upper_bound)

    @pytest.mark.parametrize("arg", ["1,2", "1;2", "1_2", "1/2", "1.2", "1~2"])
    @given(upper_bound=st.integers(min_value=3))
    def test_illegal_delimiter__fail_with_expected_error(self, arg, upper_bound):
        with pytest.raises(NatIntervalException):
            NatIntervalInclusive.from_string(arg, upper_bound=upper_bound)

    @given(
        lower_and_limit=st.tuples(st.integers(), st.integers()).filter(
            lambda p: p[1] < p[0]
        )
    )
    def test_one_sided_lower_with_samples_lt_bound__fails(
        self, lower_and_limit, legit_delim
    ):
        lower, limit = lower_and_limit
        arg = str(lower) + legit_delim
        with pytest.raises(NatIntervalException):
            NatIntervalInclusive.from_string(arg, upper_bound=limit)

    @given(lower_and_upper=nondecreasing_pair_strategy(min_value=1))
    def test_one_sided_lower_with_samples_gteq_bound__succeeds(
        self, lower_and_upper, legit_delim
    ):
        lo, upper_bound = lower_and_upper
        exp = NatIntervalInclusive(lo, upper_bound)
        arg = str(lo) + legit_delim
        obs = NatIntervalInclusive.from_string(arg, upper_bound=upper_bound)
        assert obs == exp

    @given(upper_and_limit=nondecreasing_pair_strategy(min_value=1))
    def test_one_sided_upper_with_samples_gteq_bound__succeeds(
        self, upper_and_limit, legit_delim
    ):
        upper, limit = upper_and_limit
        exp = NatIntervalInclusive(1, upper)
        arg = legit_delim + str(upper)
        obs = NatIntervalInclusive.from_string(arg, upper_bound=limit)
        assert obs == exp

    @given(
        upper_and_limit=st.tuples(
            st.integers(min_value=1), st.integers(min_value=1)
        ).filter(lambda p: p[1] < p[0])
    )
    def test_one_sided_upper_with_samples_lt_bound__uses_bound(
        self, upper_and_limit, legit_delim
    ):
        upper, limit = upper_and_limit
        exp = NatIntervalInclusive(1, limit)
        arg = legit_delim + str(upper)
        obs = NatIntervalInclusive.from_string(arg, upper_bound=limit)
        assert obs == exp

    @given(
        lower_upper_limit=st.tuples(gen_pos_int, gen_pos_int, gen_pos_int).filter(
            lambda t: t[1] < t[0] or t[2] < t[0]
        )
    )
    def test_two_sided_parse_upper_lt_lower(self, lower_upper_limit, legit_delim):
        lo, hi, lim = lower_upper_limit
        arg = str(lo) + legit_delim + str(hi)
        with pytest.raises(NatIntervalException):
            NatIntervalInclusive.from_string(arg, upper_bound=lim)

    @given(
        lo_hi_limit=st.tuples(
            st.integers(min_value=2), gen_pos_int, gen_pos_int
        ).filter(lambda t: t[2] < t[0] <= t[1])
    )
    def test_two_sided_parse_upper_gteq_lower_with_upper_limit_lt_lower(
        self, lo_hi_limit, legit_delim
    ):
        lo, hi, limit = lo_hi_limit
        arg = str(lo) + legit_delim + str(hi)
        with pytest.raises(NatIntervalException):
            NatIntervalInclusive.from_string(arg, upper_bound=limit)

    @given(
        lo_hi_limit=st.tuples(gen_pos_int, gen_pos_int, gen_pos_int).filter(
            lambda t: t[0] < t[2] < t[1]
        )
    )
    def test_two_sided_parse_upper_gteq_lower_with_upper_limit_between_lower_and_upper(
        self,
        lo_hi_limit,
        legit_delim,
    ):
        lo, hi, limit = lo_hi_limit
        exp = NatIntervalInclusive(lo, limit)
        arg = str(lo) + legit_delim + str(hi)
        obs = NatIntervalInclusive.from_string(arg, upper_bound=limit)
        assert obs == exp

    @given(
        lo_hi_upper=st.tuples(gen_pos_int, gen_pos_int, gen_pos_int).filter(
            lambda t: t[0] <= t[1] <= t[2]
        )
    )
    def test_two_sided_parse_upper_gteq_lower_with_upper_limit_gteq_upper(
        self, lo_hi_upper, legit_delim
    ):
        lo, hi, upper_bound = lo_hi_upper
        exp = NatIntervalInclusive(lo, hi)
        arg = f"{str(lo)}{legit_delim}{str(hi)}"
        obs = NatIntervalInclusive.from_string(arg, upper_bound=upper_bound)
        assert obs == exp



""" Definitions of the parser argument types """

from yacman import YAMLConfigManager


def html_range(caravel=False, min_val=0, max_val=10, step=1, value=0):
    caravel_data = YAMLConfigManager(
        {
            "element_type": "range",
            "element_args": {
                "min": min_val,
                "max": max_val,
                "step": step,
                "value": value,
            },
        }
    )
    if step < 1:

        def fun(x=None, caravel_data=caravel_data, caravel=caravel):
            return caravel_data if caravel else float(x)

    else:

        def fun(x=None, caravel_data=caravel_data, caravel=caravel):
            return caravel_data if caravel else str(x)

    return fun


def html_checkbox(caravel=False, checked=False):
    """
    Create argument for type parameter on argparse.ArgumentParser.add_argument.

    :param bool caravel: whether this is being used in the caravel context
    :param bool checked: whether to add a particular key-value entry to a
        collection used by caravel
    :return callable: argument to the type parameter of an
        argparse.ArgumentParser's add_argument method.
    """
    caravel_data = YAMLConfigManager({"element_type": "checkbox", "element_args": {}})
    if checked:
        caravel_data.add_entries({"element_args": {"checked": True}})

    def fun(x=None, caravel_data=caravel_data, caravel=caravel):
        return caravel_data if caravel else eval(x)

    return fun


def html_select(choices, caravel=False):
    """
    Create argument for type parameter on argparse.ArgumentParser.add_argument.

    :param list[object] choices: collection of valid argument provisions via
        to a particular CLI option
    :param bool caravel: whether this is being used in the caravel context
    :return callable: argument to the type parameter of an
        argparse.ArgumentParser's add_argument method.
    """
    if not isinstance(choices, list):
        raise TypeError(
            "Argument to choices parameter must be list, got {}.".format(type(choices))
        )
    caravel_data = YAMLConfigManager(
        {"element_type": "select", "element_args": {"option": choices}}
    )

    def fun(x=None, caravel_data=caravel_data, caravel=caravel):
        return caravel_data if caravel else x

    return fun



import sys

from .cli_looper import main
from .cli_divvy import main as divvy_main

if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        print("Program canceled by user!")
        sys.exit(1)



"""Tests for looper's cleaning functionality"""

import argparse
import pytest
from looper.looper import Cleaner
from looper import Project


def build_namespace(**kwargs):
    """Create an argparse namespace with given key-value pairs."""
    ns = argparse.Namespace()
    for opt, arg in kwargs.items():
        setattr(ns, opt, arg)
    return ns


DRYRUN_OR_NOT_PREVIEW = [
    pytest.param(args, preview, id=param_name)
    for args, preview, param_name in [
        (build_namespace(dry_run=False), False, "not_preview__only"),
        (build_namespace(dry_run=True), False, "dry_run__only"),
        (build_namespace(dry_run=True), True, "dry_run__and__not_preview"),
    ]
]


@pytest.mark.parametrize(["args", "preview"], DRYRUN_OR_NOT_PREVIEW)
def test_cleaner_does_not_crash(args, preview, prep_temp_pep):
    prj = Project(prep_temp_pep)
    prj._samples = []
    clean = Cleaner(prj)
    try:
        retcode = clean(args=args, preview_flag=preview)
    except Exception as e:
        pytest.fail(f"Cleaning call hit error: {e}")
    else:
        assert retcode == 0
